# -*- coding: utf-8 -*-
"""
Created by chiesa

Copyright Alpes Lasers SA, Switzerland
"""
__author__ = 'chiesa'
__copyright__ = "Copyright Alpes Lasers SA"

import requests


def stop(url='http://0.0.0.0:5000'):
    rsp = requests.post(url + '/scan/stop')
    rsp.raise_for_status()


if __name__ == '__main__':
    stop(url='http://phootonix:5000')


from .ClassAbstracts import System
from ..tools.systems_registers_tools import is_valid_channel


class InputConditioning(System):
    r"""
Input Conditioning Channel - Offset and Scaling
System ID: 0x98 through 0x9f

+----------------+------+-------------+--------------------+-------------+---------+------------------------+
| Register Name  | Id   | Type        | Unit               | Range       | Default | Comment                |
+================+======+=============+====================+=============+=========+========================+
| gain           | 0x00 | float 32-bit|                    |             | 1.0     |                        |
+----------------+------+-------------+--------------------+-------------+---------+------------------------+
| offset         | 0x01 | float 32-bit|                    |             | 0.0     |                        |
+----------------+------+-------------+--------------------+-------------+---------+------------------------+

    """

    @staticmethod
    def help():
        print(InputConditioning.__doc__)

    def __init__(self, channel: int = 0, board=None):
        self.sys_id = 0x98 | channel
        self._readonly = False

        self.gain = {'id': self.sys_id << 8 | 0x00,
                     'type': float,
                     'unit': None,
                     'range': None,
                     'default': 1,
                     'value': 1}

        self.offset = {'id': self.sys_id << 8 | 0x01,
                       'type': float,
                       'unit': None,
                       'range': None,
                       'default': 0,
                       'value': 0}
        System.__init__(self, channel, board)
        self.name = self.__class__.__name__
        if not is_valid_channel(self._channel):
            raise ValueError('Channel Range Error')

    def SetGain(self, value):
        return self.set_register('gain', value)

    def SetOffset(self, value):
        return self.set_register('offset', value)

    def GetGain(self):
        return self.get_register('gain')

    def GetOffset(self):
        return self.get_register('offset')


# -*- coding: utf-8 -*-
"""
Created by chiesa

Copyright Alpes Lasers SA, Switzerland
"""
__author__ = 'chiesa'
__copyright__ = "Copyright Alpes Lasers SA"

from pprint import PrettyPrinter

import requests


def monitor(url='http://0.0.0.0:5000'):
    rsp = requests.get(url + '/info')
    rsp.raise_for_status()
    pp = PrettyPrinter(indent=1)
    pp.pprint(rsp.json()['monitoring'])


def info(url='http://0.0.0.0:5000'):
    rsp = requests.get(url + '/info')
    rsp.raise_for_status()
    pp = PrettyPrinter(indent=1)
    pp.pprint(rsp.json())


if __name__ == '__main__':
    info(url='http://phootonix:5000')


# -*- coding: utf-8 -*-
"""
Created by gregory on 06.01.17

Copyright 2017 Alpes Lasers SA, Neuchatel, Switzerland
"""

import logging

from .base_controller import BaseController
import threading
import serial

logger = logging.getLogger(__name__)

__author__ = 'gregory'
__copyright__ = "Copyright 2016, Alpes Lasers SA"

ON = 'ON'
OFF = 'OFF'
READ = 'READ'
CHANNELS = {
    1: {ON: '1', OFF: 'Q', READ: 'A'},
    2: {ON: '2', OFF: 'W', READ: 'S'},
    3: {ON: '3', OFF: 'E', READ: 'D'},
    4: {ON: '4', OFF: 'R', READ: 'F'},
    5: {ON: '5', OFF: 'T', READ: 'G'},
    6: {ON: '6', OFF: 'Y', READ: 'H'},
    7: {ON: '7', OFF: 'U', READ: 'J'},
    8: {ON: '8', OFF: 'I', READ: 'K'},
}


class DLPController(BaseController):
    READY = 'READY'

    def __init__(self, port):
        super(DLPController, self).__init__()
        self._device = None
        self._port = port
        self._baudrate = 115200
        self._comm_lock = threading.Lock()

    def _connected_status(self):
        return DLPController.READY

    def read_channel(self, channel):
        with self._comm_lock:
            self._device.write(CHANNELS[channel][READ])
            ans = None
            while True:
                c = self._device.read()
                if not c:
                    raise Exception('DLP did not answer')
                if c not in '\r\n':
                    ans = bool(int(c))
                elif self._device.in_waiting == 0:
                    break
            return ans

    def set_channel(self, channel, value):
        with self._comm_lock:
            self._device.write(CHANNELS[channel][ON if value else OFF])

    def _connect(self):
        self._device = serial.Serial(self._port, baudrate=self._baudrate, timeout=2.0)

    def _disconnect(self):
        try:
            self._device.close()
        finally:
            self._device = None

    def _check_connection(self):
        with self._comm_lock:
            self._device.write("'")
            return self._device.read() == 'Q'



# -*- coding: utf-8 -*-
"""
Created by gregory on 17.03.16

Copyright 2016 Alpes Lasers SA, Neuchatel, Switzerland
"""

import logging
import numbers
import math

logger = logging.getLogger(__name__)

__author__ = 'gregory'
__copyright__ = "Copyright 2016, Alpes Lasers SA"


def validate_calibration(calibration):
    error = False
    if not isinstance(calibration, list):
        error = True
    else:
        for tuple in calibration:
           # if len(tuple) != 2 or not isinstance(tuple[0], numbers.Number) or tuple[0] < 0 \
           #         or not isinstance(tuple[1], numbers.Number) or tuple[1] < 0:
            if len(tuple) != 2 or not isinstance(tuple[0], numbers.Number) \
                    or not isinstance(tuple[1], numbers.Number) or tuple[1] < 0:
                error = True
                break
    if error:
        raise ValueError('The calibration must be given as a list of pairs angle, wavelength,'
                         'e.g. [(350, 1500), (351, 1512), ...])')


def _angle_delta_old(angle1, angle2):
    delta = math.fmod(angle1 - angle2, 360.0)
    if delta > 180.0:
        return delta - 360.0
    elif delta < -180.0:
        return delta + 360.0
    return delta


def _angle_delta(angle1, angle2):
    # delta = math.fmod(angle1 - angle2, 360.0)
    # if delta > 180.0:
    #     return delta - 360.0
    # elif delta < -180.0:
    #     return delta + 360.0
    # return delta
    delta = angle1 - angle2
    return delta


def _reduced_angle(angle):
    # reduced = math.fmod(angle, 360.0)
    # if reduced < 0:
    #     return reduced + 360.0
    # return reduced
    return angle


def _reduced_angle_old(angle):
    reduced = math.fmod(angle, 360.0)
    if reduced < 0:
        return reduced + 360.0
    return reduced


def angle_for_wavelength(calibration, wavelength):
    prev_angle, prev_wl = None, None
    for angle, wl in calibration:
        if prev_wl is not None:
            cpd = float(wavelength - prev_wl)
            wcd = float(wl - wavelength)
            if cpd * wcd >= 0:
                s = cpd / float(wl - prev_wl)
                return _reduced_angle(prev_angle + s * _angle_delta(angle, prev_angle))
        prev_angle, prev_wl = angle, wl
    raise ValueError('The desired wavelength is outside of the calibration range.')


def wavelength_for_angle(calibration, current_angle):
    prev_angle, prev_wl = None, None
    for angle, wl in calibration:
        if prev_angle is not None:
            cpd = _angle_delta(current_angle, prev_angle)
            acd = _angle_delta(angle, current_angle)
            if cpd * acd >= 0:
                s = cpd / _angle_delta(angle, prev_angle)
                return prev_wl + s * (wl - prev_wl)
        prev_angle, prev_wl = angle, wl
    raise ValueError('The given angle is outside of the calibration range.')


def angle_for_wavelength_old(calibration, wavelength):
    prev_angle, prev_wl = None, None
    for angle, wl in calibration:
        if prev_wl is not None:
            cpd = float(wavelength - prev_wl)
            wcd = float(wl - wavelength)
            if cpd * wcd >= 0:
                s = cpd / float(wl - prev_wl)
                return _reduced_angle_old(prev_angle + s * _angle_delta_old(angle, prev_angle))
        prev_angle, prev_wl = angle, wl
    raise ValueError('The desired wavelength is outside of the calibration range.')


def wavelength_for_angle_old(calibration, current_angle):
    prev_angle, prev_wl = None, None
    for angle, wl in calibration:
        if prev_angle is not None:
            cpd = _angle_delta_old(current_angle, prev_angle)
            acd = _angle_delta_old(angle, current_angle)
            if cpd * acd >= 0:
                s = cpd / _angle_delta_old(angle, prev_angle)
                return prev_wl + s * (wl - prev_wl)
        prev_angle, prev_wl = angle, wl
    raise ValueError('The given angle is outside of the calibration range.')

if __name__ == '__main__':
    test_calibration = [
        (309.5, 5089.73),
        (309.0, 5126.55),
        (308.5, 5163.52),
        (308.0, 5199.6),
        (307.5, 5234.88),
        (307.0, 5269.71),
        (306.5, 5305.75),
        (306.0, 5341.4),
        (305.5, 5376.71),
        (305.0, 5410.91),
        (304.5, 5444.79),
        (304.0, 5479.02),
        (303.5, 5513.56),
        (303.0, 5547.3),
        (302.5, 5576.99),
        (302.0, 5610.65),
        (301.5, 5641.1),
        (300.5, 5701.11),
        (300.0, 5730.07),
        (299.5, 5759.29),
        (299.0, 5789.03),
        (298.5, 5817.98),
        (298.0, 5847.06),
        (297.5, 5874.83),
        (297.0, 5902.25),
        (301.0, 5671.68),
        (296.5, 5929.32),
        (296.0, 5957.04),
        (295.5, 5983.46),
        (295.0, 6009.48),
        (294.5, 6034.35),
        (294.0, 6058.87),
        (293.5, 6082.37),
        (293.0, 6105.56), ]

    import numpy as np
    lines = ["angle [deg],wavelength [nm]"]
    for angle in np.arange(309.5, 293, -0.001):
         lines.append("{:.4f},{:.2f}".format(angle, wavelength_for_angle(test_calibration, angle)))
    with open("calibration.csv", 'w') as fp:
        fp.write('\n'.join(lines))
    test_calibration = list(reversed(test_calibration))
    angle = angle_for_wavelength(test_calibration, 5701.12)
    print(angle)
    print(wavelength_for_angle(test_calibration, angle))



# -*- coding: utf-8 -*-
"""
Created by gregory on 16.03.16

Copyright 2016 Alpes Lasers SA, Neuchatel, Switzerland
"""

import logging
import time
import threading

from .vcm_configuration import (configuration_dictionary, STAGE_ANGLE, update_configuration, MAXIMUM_STAGE_VELOCITY,
                                                         get_wavelength_for_angle)
from motor_apt.core2.cr1 import CR1

from .base_controller import BaseController, run_in_thread
from .dlp import DLPController

logger = logging.getLogger(__name__)
__author__ = 'gregory'
__copyright__ = "Copyright 2016, Alpes Lasers SA"


class StageBusyError(Exception):
    pass


class OutOfRangeError(Exception):
    pass


class StageController(BaseController):
    READY = 'READY'
    BUSY = 'BUSY'
    SWEEP = 'SWEEP'
    STEPWISE = 'STEPWISE'

    @property
    def has_two_axes(self):
        return bool(self._has_two_axes)

    @property
    def is_busy(self):
        return self._device is not None and self._device_status is not None and (self._device_status.moving or
                                                                                 self._device_status.homing or
                                             (self._worker_thread and self._worker_thread.is_alive()))

    def __init__(self, serial_number=None, dlp_controller=None, monitor_interval=1.0):
        super(StageController, self).__init__(monitor_interval=monitor_interval)

        self._serial_number = serial_number
        self._device_info = None
        self._device_status = None
        self._busy_status = None  # is SWEEP or STEPWISE or None
        self._device = None
        self._dlp_controller = dlp_controller

        self._sweep_data = []
        self._worker_thread = None
        self._stop_worker_event = threading.Event()
        self._has_two_axes = False

    def get_angle(self):
        if self._device_status:
            return self._device_status.position
        else:
            return 'N/A'

    def get_info(self):
        return self._device_info

    def stop_motion(self):
        self._stop_worker_event.set()
        self._worker_thread = None
        self._device.stop(wait=True, immediate=True)

    def set_current_angle(self, angle):
        update_configuration({STAGE_ANGLE: angle})
        self._device.set_encoder_counter(int(self._device.enccnt * angle))

    def move_to_angle(self, angle):
        if self.is_busy:
            raise StageBusyError()
        self._device.move_absolute(angle, wait=False)

    def jog(self, step):
        if self.is_busy:
            raise StageBusyError()
        jog_mode, step_size, min_vel, acc, max_vel, stop_mode = self._device.jog_parameters()
        if acc > self._device.max_acceleration:
            acc = self._device.max_acceleration
        if max_vel > self._device.max_velocity:
            max_vel = self._device.max_velocity
        self._device.set_jog_parameters(jog_mode=jog_mode, step_size=abs(step), acceleration=acc, max_velocity=max_vel,
                                        stop_mode=stop_mode)
        self._device.jog(direction=2 if step > 0 else 1, wait=False)

    def sweep(self, start_angle, end_angle, velocity, repetitions=1, time_interval=0.05):
        if self.is_busy:
            raise StageBusyError()
        self._stop_worker_event.clear()
        self._worker_thread = self._sweep_worker(start_angle, end_angle, velocity, repetitions, time_interval)

    def get_sweep_data(self):
        return self._sweep_data

    def stepwise(self, steps, delay_s):
        if self.is_busy:
            raise StageBusyError()
        self._stop_worker_event.clear()
        self._worker_thread = self._stepwise_worker(steps, delay_s)

    @run_in_thread
    def _sweep_worker(self, start_angle, end_angle, velocity, repetitions=1, time_interval=0.05):
        min_vel, acc, max_vel = self._device.velocity_parameters()
        if acc > self._device.max_acceleration:
            acc = self._device.max_acceleration
        max_vel = self._device.max_velocity
        try:
            self._busy_status = StageController.SWEEP
            self._sweep_data = []
            while repetitions == -1 or repetitions > 0:
                self._device.set_velocity_parameters(acc, max_vel)
                self.__notify_ramping(False, True)
                self._device.move_absolute(start_angle, wait=True)
                if self._stop_worker_event.is_set():
                    return
                self._device.set_velocity_parameters(acc, velocity)

                start_time = time.time()
                status = self._device.status()
                current_sweep = [(0, status.position, get_wavelength_for_angle(status.position, False))]
                self._device.move_absolute(end_angle, wait=False)
                self.__notify_ramping(True, False)
                moving = True
                while moving:
                    time.sleep(time_interval)
                    if self._stop_worker_event.is_set():
                        return
                    status = self._device.status()
                    moving = status.moving
                    logger.debug('Current position: {}'.format(status.position))
                    current_sweep.append((time.time() - start_time, status.position,
                                          get_wavelength_for_angle(status.position, False)))
                self.__notify_ramping(False, False)
                if repetitions > 0:
                    self._sweep_data.append(current_sweep)
                    repetitions -= 1
                else:
                    self._sweep_data = [current_sweep]
                if self._stop_worker_event.is_set():
                    return
        finally:
            self.__notify_ramping(False, False)
            self._device.set_velocity_parameters(acc, max_vel)
            self._busy_status = None

    def __notify_ramping(self, is_ramping_forward, is_ramping_backwards):
        if self._dlp_controller and self._dlp_controller.status == DLPController.READY:
            self._dlp_controller.set_channel(1, is_ramping_forward)
            self._dlp_controller.set_channel(2, is_ramping_backwards)

    @run_in_thread
    def _stepwise_worker(self, steps, delay_s):
        try:
            self._busy_status = StageController.STEPWISE
            for step in steps:
                if self._stop_worker_event.is_set():
                    break
                self._device.move_absolute(step, wait=True)
                time.sleep(delay_s)
        finally:
            self._busy_status = None

    def _connected_status(self):
        if self.is_busy:
            return self._busy_status or StageController.BUSY
        return StageController.READY

    def _disconnect(self):
        last_angle = self.get_angle()
        if last_angle != 'N/A':
            update_configuration({STAGE_ANGLE: last_angle})
        try:
            self.stop_motion()
            self._device_info = None
        finally:
            try:
                self._device.close()
            except:
                pass
            self._device = None

    def _connect(self):
        try:
            self._device = CR1(self._serial_number)
            self._device.open()
            self.set_current_angle(configuration_dictionary[STAGE_ANGLE])
            self._device.set_velocity_parameters(self._device.max_acceleration, configuration_dictionary[MAXIMUM_STAGE_VELOCITY])
        except:
            raise Exception('Cannot connect.')
        serial_number, model, hwtype, fwver, notes, hwver, modstate, numchan = self._device.info()
        self._device_info = {'serial_number': serial_number, 'model': model, 'hardware_type': hwtype,
                             'firmware_version': fwver, 'notes': notes, 'hardware_version': hwver,
                             'modification_state': modstate, 'no_of_channels': numchan}

    def _check_connection(self):
        last_angle = self.get_angle()
        try:
            self._device_status = self._device.status()
            if last_angle != 'N/A':
                update_configuration({STAGE_ANGLE: last_angle})
        except:
            # persist the last known angle if the motor connection fails.
            if last_angle != 'N/A':
                update_configuration({STAGE_ANGLE: last_angle})
            raise
        return True


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    logger.setLevel(logging.DEBUG)
    controller = StageController()
    controller.start()
    try:
        time.sleep(2)
        controller.sweep(0, 2, 2, 2, 0.05)
        while controller.is_busy:
            time.sleep(0.5)
        print(controller.get_sweep_data())
    finally:
        controller.stop()



from logging import getLogger

import numpy as np
import time
import struct
import inspect
from .tools.parsing_tools import encode, encode_vector
from .tools.command_tools import issue_command, process_simple_response, process_response, get_cmd_reg_val
from .tools.systems_registers_tools import process_registers
from .tools.definitions import (CMD_DELAY, ENDIAN, Units, Waveforms,
                                CommandID, SimpleResponses, ESCAPE_BYTE,
                                FRAME_BOUNDARY, ESCAPE_MASK)


logger = getLogger(__name__)

# Kummenberg Commands
class Command:
    r"""
    Command object from which to send Board commands and process responses.
    Note: these commands are inherited by a given board (Duck typed)

    For Board information, see help(connection.Board) or connection.Board.__doc__

    Parameters
    ----------
    board : Board
        The Board object from which to issue commands and receive responses.

    Returns
    -------
    response : str
        Firmware response string, or parse_error code string.

    Raises
    ------
    none

    Notes
    -----
    Board commands operate via duck typing. Must pass a valid Board.
    Command success may depend on _board mode (simple/pro)

    Examples
    --------
    Given an existing _board connection (bd) in simple mode, this will perform a simple handshake acknowledgement.

    >>> Command(bd).handshake()
    Handshake Successful.

    """

    def __init__(self, board=None):
        pass

    # Simple Mode Commands
    def handshake(self, force: bool = False):
        r"""
        Acknowledge communication in simple mode.

        Parameters
        ----------
        force : bool
            Will first ensure board is in correct mode before sending command.

        Returns
        -------
        response : bytes
            Firmware response string, or parse_error code string.
        """
        if force:
            if self.verbose:
                logger.debug('Forcing Board into Simple Mode: May return Errors.')
            self._simple = self.IsSimpleMode()
            self._set_internal_mode(self._simple, simple_mode_desired=True)
            self.go_simple()

        if self._simple:
            response = issue_command(self, 'Start\r\n'.encode())
            if response in SimpleResponses.values():
                response = SimpleResponses['PROCESSED_WITHOUT_ERROR']
        else:
            response = 0xffff
        process_simple_response(self, inspect.stack()[0][3].upper(), response)
        return response

    def go_pro(self, force: bool = False):
        r"""
        Switches _board into pro-mode.

        Parameters
        ----------
        force : bool
            Will first ensure board is in correct mode before sending command.

        Returns
        -------
        response : bytes
            Firmware response string, or parse_error code string.
        """
        if force:
            if self.verbose:
                logger.debug('Forcing Board into Pro Mode: May return Errors.')
            self._simple = self.IsSimpleMode()
            self._set_internal_mode(self._simple, simple_mode_desired=True)

        response = issue_command(self, 'GoPro\r\n'.encode())
        if response is SimpleResponses['UNRECOGNIZED_MESSAGE']:
            response = self.go_pro()
        if response in SimpleResponses.values() and response is not SimpleResponses['UNRECOGNIZED_MESSAGE']:
            self._simple = False
        return response

    def reset(self, force: bool = False):
        r"""
        Reset firmware and reconnect to Board. Will be in simple-mode after reset.
        Must be in Simple Mode to execute this Command!

        Parameters
        ----------
        force : bool
            Will first ensure board is in correct mode before sending command.

        Returns
        -------
        response : bytes
            Firmware response string, or parse_error code string.
        """
        if force:
            if self.verbose:
                logger.debug('Forcing Board into Simple Mode: May return Errors.')
            self._simple = self.IsSimpleMode()
            self._set_internal_mode(self._simple, simple_mode_desired=True)

        if not self._simple:
            response = 0xffff
            return response

        self.Connection._comm_lock = False
        self.Connection.send('Reset\r\n'.encode())
        self.Connection.disconnect()
        time.sleep(1.5)
        self.Connection.connect()
        self.Connection._comm_lock = True

        if self.verbose:
            logger.debug('Verifying Simple via Handshake...')
        response = self.handshake()
        if response in SimpleResponses.values():
            self._simple = True

        return response

    # Pro Mode Commands
    def generic_command(self, force: bool = False):
        r"""
        Sends generic pro-mode command.

        Parameters
        ----------
        force : bool
            Will first ensure board is in correct mode before sending command.

        Returns
        -------
        response : bytes
            Firmware response string, or parse_error code string.
        """
        cmd_id = CommandID.GENERIC_COMMAND
        register = {'id': 0, 'type': int, 'unit': None, 'range': [0, 1]}
        value = 0

        if force:
            if self.verbose:
                logger.debug('Forcing Board into Pro Mode: May return Errors.')
            self._simple = self.IsSimpleMode()
            self._set_internal_mode(self._simple, simple_mode_desired=False)

        if self._simple:
            response = 0xffff
            return process_simple_response(self, inspect.stack()[0][3].upper(), response)
        else:
            data = encode(cmd_id, value)
            response = issue_command(self, data)
            parsed_response = process_response(cmd_id, register, value, response, self.verbose)
        if parsed_response[4]:
            # error flag was true
            logger.debug('Last Message Sent : {}'.format(self.Connection.last_sent))
        return parsed_response[2]

    def get_firmware_id(self, force: bool = False):
        r"""
        Gets firmware identification.

        Parameters
        ----------
        force : bool
            Will first ensure board is in correct mode before sending command.

        Returns
        -------
        response : bytes
            Firmware response string, or parse_error code string.
        """
        cmd_id = CommandID.GET_FIRMWARE_ID
        register = {'id': 0, 'type': int, 'unit': None, 'range': [0, 1]}
        value = 0

        if force:
            if self.verbose:
                logger.debug('Forcing Board into Pro Mode: May return Errors.')
            self._simple = self.IsSimpleMode()
            self._set_internal_mode(self._simple, simple_mode_desired=False)

        if self._simple:
            response = 0xffff
            return process_simple_response(self, inspect.stack()[0][3].upper(), response)
        else:
            data = encode(cmd_id, value)
            response = issue_command(self, data)
            parsed_response = process_response(cmd_id, register, value, response, self.verbose)
        if parsed_response[4]:
            # error flag was true
            logger.debug('Last Message Sent : {}'.format(self.Connection.last_sent))
        return parsed_response[2]

    def get_status(self, force: bool = False):
        r"""
        Gets the board status information.

        Parameters
        ----------
        force : bool
            Will first ensure board is in correct mode before sending command.

        Returns
        -------
        response : bytes
            Firmware response string, or parse_error code string.
        """
        cmd_id = CommandID.GET_STATUS
        register = {'id': 0, 'type': int, 'unit': None, 'range': [0, 1]}
        value = 0

        if force:
            if self.verbose:
                logger.debug('Forcing Board into Pro Mode: May return Errors.')
            self._simple = self.IsSimpleMode()
            self._set_internal_mode(self._simple, simple_mode_desired=False)

        if self._simple:
            response = 0xffff
            return process_simple_response(self, inspect.stack()[0][3].upper(), response)
        else:
            data = encode(cmd_id, value)
            response = issue_command(self, data)
            parsed_response = process_response(cmd_id, register, value, response, self.verbose)
        if parsed_response[4]:
            # error flag was true
            logger.debug('Last Message Sent : {}'.format(self.Connection.last_sent))
        return parsed_response[2]

    def start_self_test(self, force: bool = False):
        r"""
        Begins board self-test.

        Parameters
        ----------
        force : bool
            Will first ensure board is in correct mode before sending command.

        Returns
        -------
        response : bytes
            Firmware response string, or parse_error code string.
        """
        cmd_id = CommandID.START_SELF_TEST
        register = {'id': 0, 'type': int, 'unit': None, 'range': [0, 1]}
        value = 0

        if force:
            if self.verbose:
                logger.debug('Forcing Board into Pro Mode: May return Errors.')
            self._simple = self.IsSimpleMode()
            self._set_internal_mode(self._simple, simple_mode_desired=False)

        if self._simple:
            response = 0xffff
            return process_simple_response(self, inspect.stack()[0][3].upper(), response)
        else:
            data = encode(cmd_id, value)
            response = issue_command(self, data)
            parsed_response = process_response(cmd_id, register, value, response, self.verbose)
        if parsed_response[4]:
            # error flag was true
            logger.debug('Last Message Sent : {}'.format(self.Connection.last_sent))
        return parsed_response[2]

    def load_snapshot(self, memory_id: int, force: bool = False):
        r"""
        Loads snapshot at given memory location, may not work if id is invalid.
        For more information, see: SnapshotManager in registers.

        Parameters
        ----------
        memory_id : int
            Memory location to load.
        force : bool
            Will first ensure board is in correct mode before sending command.

        Returns
        -------
        response : bytes
            Firmware response string, or parse_error code string.
        """
        cmd_id = CommandID.LOAD_CONFIG
        register = {'id': 0, 'type': int, 'unit': None, 'range': [0, 7]}
        value = memory_id

        if force:
            if self.verbose:
                logger.debug('Forcing Board into Pro Mode: May return Errors.')
            self._simple = self.IsSimpleMode()
            self._set_internal_mode(self._simple, simple_mode_desired=False)

        data = encode(cmd_id, value)
        response = issue_command(self, data)

        parsed_response = process_response(cmd_id, register, value, response, self.verbose)
        if parsed_response[4]:
            # error flag was true
            logger.debug('Last Message Sent : {}'.format(self.Connection.last_sent))
        return parsed_response[2]

    def save_snapshot(self, memory_id: int, force: bool = False):
        r"""
        Saves snapshot at given memory location, may not work if id is invalid. (i.e. location 0 is readonly)
        For more information, see: SnapshotManager in registers.

        Parameters
        ----------
        memory_id : int
            Memory location to save.
        force : bool
            Will first ensure board is in correct mode before sending command.

        Returns
        -------
        response : bytes
            Firmware response string, or parse_error code string.
        """
        cmd_id = CommandID.STORE_CONFIG
        register = {'id': 0, 'type': int, 'unit': None, 'range': [0, 7]}
        value = memory_id

        if force:
            if self.verbose:
                logger.debug('Forcing Board into Pro Mode: May return Errors.')
            self._simple = self.IsSimpleMode()
            self._set_internal_mode(self._simple, simple_mode_desired=False)

        data = encode(cmd_id, value)
        response = issue_command(self, data, cmd_delay=1)
        parsed_response = process_response(cmd_id, register, value, response, self.verbose)
        if parsed_response[4]:
            # error flag was true
            logger.debug('Last Message Sent : {}'.format(self.Connection.last_sent))
        return parsed_response[2]

    def go_simple(self, force: bool = False):
        r"""
        Switches _board into simple-mode.

        Parameters
        ----------
        force : bool
            Will first ensure board is in correct mode before sending command.

        Returns
        -------
        response : bytes
            Firmware response string, or parse_error code string.
        """
        cmd_id = CommandID.SET_COMM_MODE
        register = {'id': 0, 'type': int, 'unit': None, 'range': [0, 1]}
        value = 0
        data = encode(CommandID.SET_COMM_MODE, 0)

        if force:
            if self.verbose:
                logger.debug('Forcing Board into Pro Mode: May return Errors.')
            self._simple = self.IsSimpleMode()
            self._set_internal_mode(self._simple, simple_mode_desired=False)

        if self._simple:
            response = 0xffff
            return process_simple_response(self, 'Go Simple', response)
        else:
            response = issue_command(self, data)
            parsed_response = process_response(cmd_id, register, value,
                                               response, self.verbose) if not self.simulated else (None,)*5

        if parsed_response[1] == CommandID.SET_COMM_MODE or force or self.simulated:
            self._simple = True
            if self.verbose:
                logger.debug('Verifying Simple via Handshake...')
            simple_mode_check = self.handshake()
            if not simple_mode_check in SimpleResponses.values():
                # Try again once
                simple_mode_check = self.handshake()
                if not simple_mode_check in SimpleResponses.values():
                    self._simple = False

        # return slave_id, response_id, response_data, crc
        return response

    def set_value(self, register: dict or list or int, value, cmd_delay: float = None):
        r"""
        Sets the given registers to desired value.

        Parameters
        ----------
        register : dict or list of dicts or int or list of ints
            Register dictionary corresponding to desired registers.{'id': registers number, 'type': data type, ...}
            Further, a raw registers id can be passed, in which case no validation will occur (COULD CAUSE ERRORS).
        value : varies, can also be list of values
            The type and range of value is dependent on the registers being addressed, see 'type' element above.
        cmd_delay : float
            Amount of additional time in seconds to wait between send / receive.

        Returns
        -------
        str
            Firmware response, contains acknowledgement or parse_error information.

        Raises
        ------
        none

        See Also
        --------
        Register IDs, value type and range can be found in the registers module.

        Examples
        --------
        In this example, the same _board (con) is used and the X channel OF is set to 0 over USB/UART
        Note: system and registers name must be known, see registers module for more info.

        >>> from registers.generic_registers import StaticInput
        >>> from src.main.python.optoKummenberg.connections import Board
        >>> board = Board()
        >>> registers = StaticInput(0).of
        >>> board.set_value(registers, 0.0)

        """
        # get hex information from registers & values
        cmd_reg_val = get_cmd_reg_val(command=CommandID.SET_VALUE, register=register, value=value)
        if cmd_reg_val == (None, None, None):
            return [None]
        else:
            # perform encode, issue, process
            data = encode(*cmd_reg_val)
            response = issue_command(self, data, cmd_delay=cmd_delay)
            parsed_response = process_response(cmd_reg_val[0], register, value, response, self.verbose)

            if parsed_response[4]:
                # error flag was true
                logger.debug('Last Message Sent : {}'.format(self.Connection.last_sent))
            return parsed_response[2]

    def get_value(self, register: dict or list):
        r"""
        Gets the given registers value.

        Parameters
        ----------
        register_record : dict or list
            Register dictionary fields corresponding to register.{'id': registers number, 'type': data type, ...}
        set_internal : bool
            Determines whether to set the returned value internally. (recommended)

        Returns
        -------
        str
            Firmware response, contains registers value, along with acknowledgement or parse_error information.

        Raises
        ------
        none

        See Also
        --------
        Register IDs, value type and range can be found in the registers module.

        Examples
        --------
        In this example, the same _board (_board) is used and the X channel OF is set to 0.0 over USB/UART
        Note: system and registers name must be known, see registers module for more info.

        >>> board.initialize_channel(0)
        >>> register_record = board.Channel_0.StaticInput.of
        >>> board.get_value(register_record)
        Sending: b'~\x00\x11\x02P\x01\x00\x00~'
        Received: b'~\x00\x11\x04\x00\x00\x00\x00\xf2\x81~'
        Command: [GET_VALUE]. Register: [0x5001]. Value: [[0.0]].
        Data Received: [0.0]
        (0, 17, [0.0], [242, 129])

        """
        # get hex information from registers & values
        cmd_reg_val = get_cmd_reg_val(command=CommandID.GET_VALUE, register=register, value=None)

        # synthetic value used for simulated connection
        value = None
        if self.Connection._port is None:
            if isinstance(register, list):
                value = [item['value'] for item in register]
            elif isinstance(register, dict):
                value = register['value']
        if cmd_reg_val == (None, None, None):
            return [None]
        else:
            # perform encode, issue, process
            data = encode(*cmd_reg_val)
            response = issue_command(self, data)
            parsed_response = process_response(cmd_reg_val[0], register, value, response, self.verbose)
            if parsed_response[4]:
                # error flag was true
                logger.debug('Last Message Sent : {}'.format(self.Connection.last_sent))
            return parsed_response[2]

    def set_vector(self, register: dict or int, index: int, vector, cmd_delay=None):
        r"""
        Sets the given register vector to desired value.

        Parameters
        ----------
        register : dict or list of dicts or int or list of ints
            Register dictionary corresponding to desired registers.{'id': registers number, 'type': data type, ...}
            Further, a raw registers id can be passed, in which case no validation will occur (COULD CAUSE ERRORS).
        index : int
            The location to store the given vector.
        value : varies, can also be list of values
            The type and range of value is dependent on the registers being addressed, see 'type' element above.
        cmd_delay : float
            How long to wait for response in secs. By default is set to registers.CMD_DELAY


        Returns
        -------
        str
            Firmware response, contains acknowledgement or parse_error information.

        Raises
        ------
        none

        See Also
        --------
        Register IDs, value type and range can be found in the registers module.

        Examples
        --------
        In this example, the same _board (con) is used and the X channel OF is set to 0 over USB/UART
        Note: system and registers name must be known, see registers module for more info.

        >>> from registers.generic_registers import VectorPatternMemory
        >>> from src.main.python.optoKummenberg.connections import Board
        >>> board = Board()
        >>> register = {'id': VectorPatternMemory().sys_id << 8, 'type': float}  # could also send register = 0x2600
        >>> vector = [0, 0.1, 0.2, 0.3]
        >>> board.set_vector(register, 0, vector)

        """
        if cmd_delay is None:
            cmd_delay = CMD_DELAY

        # perform encode, issue, process
        byte_string = encode_vector(register=register, index=index, vector=vector)
        response = issue_command(self, byte_string, cmd_delay=cmd_delay)
        parsed_response = process_response(command_id_sent=CommandID.SET_VECTOR,
                                           register=register, value=vector, response=response,
                                           verbose=self.verbose)
        if parsed_response[4]:
            # error flag was true
            logger.debug('Last Message Sent : {}'.format(self.Connection.last_sent))
        return parsed_response[2]

    def get_vector(self, register: dict or int, index: int, count: int):
        r"""
        Gets the given register vector.

        Parameters
        ----------
        register : int
            The system location of the desired vector.
        index : int
            The location of the vector to begin reading from.
        count :
            The number of vector elements, where each elements is 4 bytes if float/int or 1 byte otherwise.

        Returns
        -------
        str
            Firmware response, contains registers value, along with acknowledgement or parse_error information.

        Raises
        ------
        none

        See Also
        --------
        Register IDs, value type and range can be found in the registers module.

        Examples
        --------
        In this example, the same _board (_board) is used and the X channel OF is set to 0.0 over USB/UART
        Note: system and registers name must be known, see registers module for more info.

        >>> board.initialize_channel(0)
        >>> register_record = board.Channel_0.StaticInput.of
        >>> board.get_value(register_record)
        Sending: b'~\x00\x11\x02P\x01\x00\x00~'
        Received: b'~\x00\x11\x04\x00\x00\x00\x00\xf2\x81~'
        Command: [GET_VALUE]. Register: [0x5001]. Value: [[0.0]].
        Data Received: [0.0]
        (0, 17, [0.0], [242, 129])

        """
        index, count = index*4, count*4
        cmd_id = CommandID.GET_VECTOR
        if type(register) is int:
            register_id = register
        else:
            register_id = register['id']
            if register['type'] is not float:
                index, count = index//4, count//4

        # encode, issue, process
        byte_string = encode(command_id=CommandID.GET_VECTOR, register_id=register_id, data=[index, count])
        response = issue_command(board=self, command_bytes=byte_string, cmd_delay=CMD_DELAY)
        # parsed_response = process_response(cmd_id, register, count, response, self.verbose)
        parsed_response = process_response(command_id_sent=cmd_id,
                                           register=register, value=None, response=response,
                                           verbose=self.verbose)
        if parsed_response[4]:
            # error flag was true
            logger.debug('Last Message Sent : {}'.format(self.Connection.last_sent))
        return parsed_response[2]

    def disconnect(self):
        r"""
        Severs the serial connection, flushing input/output buffers.
        """
        self.Connection.disconnect()



import struct
import sys
if sys.platform.startswith('win'):
    print("Spidev is Linux-Only, thus SPI features are disabled.")
else:
    try:
        import spidev
    except ImportError:
        print("Unable to import spidev! Try: pip install spidev")
import time


class SPI:
    def __init__(self, bus: int = 0, device: int = 0):
        self._bus = bus
        self._device = device
        self._spi_comm = spidev.SpiDev()
        self.connect()
        self._spi_comm.max_speed_hz = 125000000//16
        self._spi_comm.mode = 0b01
        self._spi_comm.lsbfirst = False
        self.x_readback = 0
        self.y_readback = 0
        self.reg_value = 0
        self.status = None

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()

    def set_values(self, system1, register1, system2, register2, data1, data2, data_type='>f'):
        self.send(system1, register1, system2, register2, data1, data2, data_type)
        return self.receive_w()

    def get_values(self, system, register, unit_type):
        resp = self._spi_comm.xfer([0x00, system, register, 0, 0, 0, 0, 0, 0, 0, 0, 0])
        return self.receive_r(resp, unit_type)

    def send(self,  system1, register1, system2, register2, data1, data2, data_type='>f'):
        byte_string = struct.pack('>B', 0) + \
                      struct.pack('>B', 1) + \
                      struct.pack('>B', system1) + \
                      struct.pack('>B', register1) + \
                      struct.pack('>B', system2) + \
                      struct.pack('>B', register2) + \
                      struct.pack(data_type, data1) + \
                      struct.pack(data_type, data2)
        print(byte_string.hex())
        return self._spi_comm.xfer(byte_string)

    # TODO: Test changing pointer location
    def receive_w(self, number_of_bytes: int = 12):
        response = self._spi_comm.readbytes(number_of_bytes)
        status = format(response[0], '08b')
        status_flags = {
            'led_enabled': status[0] is True,
            'linear_drive_enabled': status[1] is True,
            'datalogging_running': status[2] is True,
            'logging_ram_full': status[3] is True,
            'conversation_error': status[4] is True,
            'overtemp_shutdown': status[5] is True,
            'current_b_overlimit': status[6] is True,
            'current_a_overlimit': status[7] is True
        }
        self.status = status_flags
        self.x_readback = struct.unpack('>f', bytes(response[4:8]))[0]
        self.y_readback = struct.unpack('>f', bytes(response[8:12]))[0]

        return status_flags, self.x_readback, self.y_readback

    def receive_r(self, unit_type, number_of_bytes: int = 12):
        response = self._spi_comm.readbytes(number_of_bytes)
        status = format(response[0], '08b')
        status_flags = {
            'led_enabled': status[0] is True,
            'linear_drive_enabled': status[1] is True,
            'datalogging_running': status[2] is True,
            'logging_ram_full': status[3] is True,
            'conversation_error': status[4] is True,
            'overtemp_shutdown': status[5] is True,
            'current_b_overlimit': status[6] is True,
            'current_a_overlimit': status[7] is True
        }
        self.reg_value = struct.unpack(unit_type, bytes(response[0:4]))[0]
        self.x_readback = struct.unpack('>f', bytes(response[4:8]))[0]
        self.y_readback = struct.unpack('>f', bytes(response[8:12]))[0]

        return self.reg_value, self.x_readback, self.y_readback

    def connect(self):
        self._spi_comm.open(self._bus, self._device)

    def disconnect(self):
        self._spi_comm.close()



import numpy

VCM_PORT = '/dev/tty.ec_vcm_main'
EXT_VCM_PORT = '/dev/tty.ec_vcm_ext'
LAST_POSITION = 1
LAST_CAVITY = 1
CAVITY1_INDEX = 'ec1'
CAVITY2_INDEX = 'ec2'
CAVITY3_INDEX = 'ec3'
CAVITY4_INDEX = 'ec4'
CAVITY1_URL = 'http://0.0.0.0:5001/ec/api/v1.0'
CAVITY2_URL = 'http://0.0.0.0:5002/ec/api/v1.0'
CAVITY3_URL = 'http://0.0.0.0:5003/ec/api/v1.0'
CAVITY4_URL = 'http://0.0.0.0:5004/ec/api/v1.0'

ADC_CHANNEL = 0
REF = 5.08  # Modify according to actual voltage (read voltage) external AVDD and AVSS (Default), or internal 2.5V
configuration_dictionary = {
    CAVITY1_INDEX: {'wavelength': [1037.2, 1353.6], 'x_position': -18.65, 'y_position': 0.5,
                    's2params': {'pulse_period': 6000, 'pulse_width': 300,
                                 'pulsing_mode': 'internal', 'applied_voltage': 18.6}, 'temperature_setpoint': 18},
    CAVITY2_INDEX: {'wavelength': [846.0, 1037.1], 'x_position': -16.75, 'y_position': -0.65,
                    's2params': {'pulse_period': 6000, 'pulse_width': 300,
                                 'pulsing_mode': 'internal', 'applied_voltage': 19.3}, 'temperature_setpoint': 18},
    CAVITY3_INDEX: {'wavelength': [1600.8, 1813.9], 'x_position': -14.35, 'y_position': -0.1,
                    's2params': {'pulse_period': 6000, 'pulse_width': 300,
                                 'pulsing_mode': 'internal', 'applied_voltage': 15.4}, 'temperature_setpoint': 18},
    CAVITY4_INDEX: {'wavelength': [1354, 1600], 'x_position': -5.5, 'y_position': -1.5,
                    's2params': {'pulse_period': 6000, 'pulse_width': 300,
                                 'pulsing_mode': 'internal', 'applied_voltage': 17.2}, 'temperature_setpoint': 18}
}

WL_INCREMENT = 1.0


def get_scanning_wavelengths(wl_step=WL_INCREMENT):
    intervals = [configuration_dictionary[CAVITY1_INDEX]['wavelength'],
                 configuration_dictionary[CAVITY2_INDEX]['wavelength'],
                 configuration_dictionary[CAVITY3_INDEX]['wavelength'],
                 configuration_dictionary[CAVITY4_INDEX]['wavelength']]
    min_wl = min(x[0] for x in intervals)
    max_wl = max(x[1] for x in intervals)
    return numpy.arange(min_wl, max_wl, wl_step).tolist()


if __name__ == '__main__':
    print(get_scanning_wavelengths())



# -*- coding: utf-8 -*-
"""
Created by gregory on 16.03.16

Copyright 2016 Alpes Lasers SA, Neuchatel, Switzerland
"""

import logging
import threading
import functools

logger = logging.getLogger(__name__)

__author__ = 'gregory'
__copyright__ = "Copyright 2016, Alpes Lasers SA"


def run_in_thread(func):
    """Decorator to run the func in a thread."""
    @functools.wraps(func)
    def threaded_func(*args, **kwargs):
        thread = threading.Thread(target=func, name=func.__name__, args=args, kwargs=kwargs)
        thread.setDaemon(True)
        thread.start()
        return thread

    return threaded_func


class BaseController(object):
    """Base class for controllers. Provides the connection logic, with several abstract protected methods that must be
    implemented in subclasses.
    """
    DISCONNECTED = 'DISCONNECTED'
    CONNECTION_ERROR = 'CONNECTION_ERROR'
    CONNECTED = 'CONNECTED'

    @property
    def status(self):
        if self.__connection_status == BaseController.CONNECTED:
            return self._connected_status()
        return self.__connection_status

    def __init__(self, monitor_interval=5.0):
        self.monitor_interval = monitor_interval  # connection monitoring interval in seconds
        self.connection_interval = 5.0  # connection retry interval.

        self.__connection_status = BaseController.DISCONNECTED
        # Private threading variables used for connection, monitoring and stopping.
        self.__connection_lock = threading.RLock()
        self.__stop_event = threading.Event()
        self.__stop_event.set()

    def start(self):
        # First command called
        """Starts the controller. It first tries to connect at regular intervals until it succeeds. Once connected, it
        monitors the connection status. It tries to reconnect if the connection is lost."""
        with self.__connection_lock:
            if self.__stop_event.is_set():
                self.__stop_event.clear()
                self.__connect_loop()

    def stop(self):
        """Stops the controller. It disconnects if needed."""
        with self.__connection_lock:
            if not self.__stop_event.is_set():
                self.__stop_event.set()
                self.__connection_status = BaseController.DISCONNECTED
                try:
                    self._disconnect()
                    logger.debug('Successfully disconnected [{}]'.format(self.__class__.__name__))
                except Exception:
                    logger.debug('Error while disconnecting [{}]'.format(self.__class__.__name__), exc_info=1)

    @run_in_thread
    def __monitor_connection(self):
        while not self.__stop_event.wait(self.monitor_interval):
            connection_ok = False
            try:
                connection_ok = self._check_connection()
            except Exception:
                logger.error('Error while checking connection [{}]'.format(self.__class__.__name__), exc_info=1)
            if not connection_ok:
                try:
                    self._disconnect()
                except Exception:
                    pass
                self.__connection_status = BaseController.CONNECTION_ERROR
                self.__connect_loop()
                return

    @run_in_thread
    def __connect_loop(self):
        while True:
            with self.__connection_lock:
                if not self.__stop_event.is_set():
                    try:
                        self._connect()
                        self._check_connection()
                        self.__connection_status = BaseController.CONNECTED
                        logger.debug('Connection successful [{}]'.format(self.__class__.__name__))
                        self.__monitor_connection()
                        return
                    except Exception as e:
                        logger.info(e)
                        if self.__connection_status != BaseController.CONNECTION_ERROR:
                            logger.debug('Could not connect [{}]'.format(self.__class__.__name__), exc_info=1)
                            self.__connection_status = BaseController.CONNECTION_ERROR
            if self.__stop_event.wait(self.connection_interval):
                return

    def _connected_status(self):
        """Protected method implemented by concrete controllers. It is to be called automatically by the base controller
        methods. It must return a status specific to the controller when it is connected."""
        return 'READY'

    def _connect(self):
        """Protected method implemented by concrete controllers. It is to be called automatically by the base controller
        methods. It must raise an Exception if the connection fails."""
        raise NotImplementedError()

    def _disconnect(self):
        """Protected method implemented by concrete controllers. It is to be called automatically by the base controller
        methods. Exceptions are logged, but otherwise ignored."""
        raise NotImplementedError()

    def _check_connection(self):
        """Protected method implemented by concrete controllers. It is to be called automatically by the base controller
        methods. It must return True if the connection is OK, False otherwise or raise an Exception."""
        raise NotImplementedError()


if __name__ == '__main__':
    # playground -- nothing relevant.
    logging.basicConfig(level=logging.DEBUG)
    controller = BaseController()
    controller._connect = lambda: True
    controller._disconnect = lambda: True
    controller._check_connection = lambda: True
    controller.connection_interval=1.0
    controller.monitor_interval = 1.0
    controller.start()
    controller.stop()
    import time
    time.sleep(5.0)
    controller.stop()
    time.sleep(5.0)



# -*- coding: utf-8 -*-
"""
Created by chiesa

Copyright Alpes Lasers SA, Switzerland
"""
__author__ = 'chiesa'
__copyright__ = "Copyright Alpes Lasers SA"

import logging
import time
from time import sleep
import requests
from requests.auth import HTTPBasicAuth

from phootonics_controller.base_controllers.config import configuration_dictionary
from phootonics_controller.utils.vcm_resources.base_controller import BaseController


target_wavelength = 1000


class ECController(BaseController):
    RESPONDING = 'RESPONDING'
    NOT_RESPONDING = 'NOT_RESPONDING'
    READY = 'READY'
    NOT_READY = 'NOT_READY'

    def __init__(self, ec_url, index, monitor_interval):
        super(ECController, self).__init__(monitor_interval=monitor_interval)
        self._driver = None
        self.index = index
        self.s2params = configuration_dictionary[self.index]['s2params']
        self.temperature = configuration_dictionary[self.index]['temperature_setpoint']
        self._device_status = None
        self._ec_client_url = ec_url
        self._stage_responsive = None
        self._stage_status = None
        self._pulser_responsive = None
        self._pulser_status = None
        self._tc_responsive = None
        self._tc_status = None

    def _connected_status(self):
        if self._stage_responsive and self._pulser_responsive and self._tc_responsive:
            return ECController.RESPONDING
        else:
            return ECController.NOT_RESPONDING

    def _check_connection(self):
        try:
            info_xc = self._driver.info_xc()
            self._pulser_status = info_xc['response']['data']['pulser']['status']
            self._tc_status = info_xc['response']['data']['tc']['status']
            self._stage_status = info_xc['response']['data']['stage']['status']
            if self._pulser_status == -1 or self._pulser_status == 2000 or self._pulser_status == 2001:
                self._pulser_responsive = False
                logging.error('pulser unresponsive: {}'.format(self._pulser_status))
            if self._tc_status == -1 or self._tc_status == 4002:
                self._tc_responsive = False
                logging.error('tc unresponsive: {}'.format(self._tc_status))
            if self._stage_status == -1 or self._stage_status == 3000 or self._stage_status == 3003 or self._stage_status == 3004:
                self._stage_responsive = False
                logging.error('stage unresponsive: {}'.format(self._stage_status))
            if self._pulser_responsive == False or self._stage_responsive == False or self._tc_responsive == False:
                logging.error('Some devices failed the connection check')
                raise
        except Exception as e:
            logging.error('connection error {}'.format(e))
            raise
        self._pulser_responsive = True
        self._tc_responsive = True
        self._stage_responsive = True
        return True

    def _connect(self):
        if self._driver is None:
            self._driver = ECClient(base_url=self._ec_client_url)

    def _disconnect(self):
        self._driver = None

    def is_temp_stable(self):
        temperature_stable = self.system_info()['response']['data']['tc']['stability']
        if temperature_stable == 'True':
            return True
        else:
            return False

    def xc_ready_to_action(self):
        try:
            if self._pulser_status in [2003, 2004] and self._tc_status == 4000 and self._stage_status == 3001 and self.is_temp_stable():
                return ECController.READY
            else:
                return ECController.NOT_READY
        except Exception as e:
            logging.exception(e)
            return ECController.NOT_READY

    def get_absolute_position(self, rel_position):
        return configuration_dictionary[self.index]['x_position'] + rel_position

    def get_absolute_position_y(self, rel_position):
        return configuration_dictionary[self.index]['y_position'] + rel_position

    def get_rel_position(self, abs_position):
        return abs_position - configuration_dictionary[self.index]['x_position']

    def get_rel_position_y(self, abs_position):
        return abs_position - configuration_dictionary[self.index]['y_position']

    def shutdown(self):
        try:
            self._driver.shutdown_xc()
        except Exception as e:
            logging.exception('shutdown error {}'.format(e))

    def system_info(self):
        return self._driver.info_xc()

    def stop_pulsing(self):
        try:
            self._driver.shut_down_s2()
        except Exception as e:
            logging.error('could not turn pulsing OFF {}'.format(e))

    def get_wavelength(self):
        try:
            return self._driver.get_wavelength()
        except Exception as e:
            logging.error('Cannot read wavelength {}'.format(e))

    def activate_s2(self):
        try:
            logging.info('Stabilizing temperature {}'.format(self.temperature))
            self._driver.set_controlled_temperature(self.temperature)
            logging.info('Set S-2 settings {}'.format(self.s2params))
            self._driver.s2_set_settings(self.s2params)
        except Exception as e:
            logging.error('Error in activating S2 {}'.format(e))

    def move_to_wavelength(self, wl, timeout=0):
        self._driver.move_to_wavelength(wl, timeout)


class MockECController(BaseController):

    RESPONDING = 'RESPONDING'
    NOT_RESPONDING = 'NOT_RESPONDING'
    READY = 'READY'
    NOT_READY = 'NOT_READY'

    def __init__(self, index, monitor_interval):
        super(MockECController, self).__init__(monitor_interval=monitor_interval)
        self._driver = None
        self.index = index
        self.s2params = configuration_dictionary[self.index]['s2params']
        self.temperature = configuration_dictionary[self.index]['temperature_setpoint']
        self._device_status = None
        self._stage_responsive = None
        self._stage_status = None
        self._pulser_responsive = None
        self._pulser_status = None
        self._tc_responsive = None
        self._tc_status = None

    def _connected_status(self):
        return ECController.RESPONDING

    def _check_connection(self):
        self._pulser_responsive = True
        self._tc_responsive = True
        self._stage_responsive = True
        return True

    def _connect(self):
        pass

    def _disconnect(self):
        pass

    def is_temp_stable(self):
        return True

    def xc_ready_to_action(self):
        return ECController.READY

    def get_absolute_position(self, rel_position):
        return 0.0

    def get_absolute_position_y(self, rel_position):
        return 0.0

    def get_rel_position(self, abs_position):
        return 0.0

    def get_rel_position_y(self, abs_position):
        return 0.0

    def shutdown(self):
        pass

    def system_info(self):
        return {'response': {'data': {}}}

    def stop_pulsing(self):
        pass

    def get_wavelength(self):
        return 1.0

    def activate_s2(self):
        pass

    def move_to_wavelength(self, wl, timeout=20):
        pass



class MotorStabilizationTimeout(Exception):
    pass


class TemperatureStabilizationTimeout(Exception):
    pass


class ECClient:
    def __init__(self, base_url,
                 temperature_tolerance=0.5,
                 temperature_stab_time=10.0,
                 temperature_stab_timeout=180.0):
        self.baseUrl = base_url
        self.auth = HTTPBasicAuth(username='admin1',
                                  password='12345')
        self.requestTimeout = 10
        self.wlTolerance = 0.1
        self.tempTol = temperature_tolerance
        self.stabTime = temperature_stab_time
        self.stabTimeout = temperature_stab_timeout

    def _get(self, path, params=None, timeout=None):
        rsp = requests.get(self.baseUrl + path, params=params, auth=self.auth, timeout=timeout or self.requestTimeout)
        rsp.raise_for_status()
        return rsp.json()

    def _post(self, path, params=None, timeout=None):
        rsp = requests.post(self.baseUrl + path, json=params, auth=self.auth, timeout=timeout or self.requestTimeout)
        rsp.raise_for_status()
        return rsp.json()

    def _is_wl_within_tolerance(self, wl_target, wl_now):
        return abs(wl_now - wl_target) < self.wlTolerance

    def move_to_wavelength(self, wl, timeout=0):
        self._post('/stage/wavelength', params={'wavelength': wl})
        if timeout == 0:
            return
        start_time = time.time()
        while True:
            status = self._get('/stage/status')['response']['code']
            if status == 3001:
                break
            if time.time() - start_time > timeout:
                raise MotorStabilizationTimeout
            time.sleep(max(timeout/100.0, 0.05))

    def stepwise_move(self, positions, delay_s):
        self._post('/stage/irrstepwise_move_wl', params=dict(positions = positions, delay_s = delay_s))

    def sweep_angle(self, amplitude, frequency, offset, shape):
        self._post('/stage/sweep_a', params=dict(amplitude = amplitude, frequency = frequency, offset = offset, shape = shape))

    def s2_set_settings(self, settings):
        pp = {k: v for k, v in settings.items() if k in ['pulse_period', 'pulse_width']}
        if pp:
            self._post('/driver/pulse_parameters', params=pp)
        if 'pulsing_mode' in settings:
            self._post('/driver/pulse_mode', params=dict(pulse_mode=settings['pulsing_mode']))
        if 'applied_voltage' in settings:
            self._post('/driver/voltage', params=dict(output_voltage=settings['applied_voltage']))

    def s2_get_settings(self):
        return

    def s2_measure(self):
        v = self._get('/driver/voltage')['response']['data']
        c = self._get('/driver/current')['response']['data']
        return {'measured_current': c['output_current'],
                'measured_voltage': v['output_voltage']}

    def wait_instrument_ready(self, timeout):
        time.sleep(timeout)

    def shut_down_s2(self):
        self.s2_set_settings(dict(applied_voltage=0.0))
        self.wait_instrument_ready(1)
        self.s2_set_settings(dict(pulsing_mode='off'))

    def get_temperature(self):
        rsp = self._get('/tc/temperature')['response']['data']
        return rsp['temperature']

    def get_wavelength(self):
        wl_now = self._get('/stage/wavelength')['response']['data']['wavelength']
        return wl_now

    def set_temperature(self, temp):
        self._post('/tc/temperature', params=dict(temperature=temp))

    def set_controlled_temperature(self, temperature):
        self.set_temperature(temperature)
        self.check_temperature_stable(temperature)

    def check_temperature_stable(self, temperature):
        self._stopCheckTempStable = False
        start = time.time()
        timeout = self.stabTimeout
        read_time_step = self.stabTime / 10.0
        stab_time = 0.0
        while True:
            if self._stopCheckTempStable:
                return
            if abs(self.get_temperature() - temperature) < self.tempTol:
                stab_time += read_time_step
            else:
                stab_time = 0.0
            if stab_time > self.stabTime:
                return
            if time.time() - start > timeout:
                raise TemperatureStabilizationTimeout
            time.sleep(read_time_step)

    def initialize_xc(self):
        self._post('/system/init')

    def info_xc(self):
        return self._get('/system/all')

    def shutdown_xc(self):
        self._post('/system/shutdown')

    def power_off(self):
        self._post('/system/power_off')


if __name__ == '__main__':
    controller = ECController('http://0.0.0.0:5001/ec/api/v1.0', monitor_interval=1)
    controller.start()
    time.sleep(1)
    try:
        print(controller.system_info())
    except Exception as e:
        print(e)
    finally:
        controller.stop()









# -*- coding: utf-8 -*-
"""
Initialize the logger

Created on July 22, 2021

Copyright Alpes Lasers SA, Neuchatel, Switzerland, 2021

@author: olgare
"""
import logging
from pkg_resources import DistributionNotFound
import pkg_resources
from distutils.version import StrictVersion

pkg = "phootonics_controller"
try:
    version = pkg_resources.get_distribution(pkg).version
    try:
        StrictVersion(version)
    except ValueError as e:
        version = 'devel'
except DistributionNotFound:
    version = "devel"



# -*- coding: utf-8 -*-
"""
Created by chiesa

Copyright Alpes Lasers SA, Switzerland
"""
__author__ = 'chiesa'
__copyright__ = "Copyright Alpes Lasers SA"

import logging
from time import sleep

from phootonics_controller.base_controllers.config import CAVITY1_INDEX
from phootonics_controller.base_controllers.xc_controller import ECController

if __name__ == '__main__':
    ecc = ECController('http://phootonix:5003/ec/api/v1.0', index=CAVITY1_INDEX, monitor_interval=0.1)
    ecc.start()
    while True:
        sleep(0.2)
        try:
            print(ecc.xc_ready_to_action())
        except Exception as e:
            logging.exception(e)




from enum import IntEnum

# Kummenberg Internal Definitions
from phootonics_controller.utils.MR_E_2_Python_SDK.optoKummenberg.tools.definitions import *

# MRE2 Specific Definitions
MRE2_CHANNEL_CNT = 2



# -*- coding: utf-8 -*-
"""
Created by gregory on 16.03.16

Copyright 2016 Alpes Lasers SA, Neuchatel, Switzerland
"""

import logging
import time
import threading

from phootonics_controller.utils.vcm_resources.base_controller import BaseController, run_in_thread
from phootonics_controller.utils.vcm_resources.vcm_configuration import (configuration_dictionary, STAGE_ANGLE)

logger = logging.getLogger(__name__)
__author__ = 'gregory'
__copyright__ = "Copyright 2016, Alpes Lasers SA"



class StageBusyError(Exception):
    pass


class OutOfRangeError(Exception):
    pass


class MockStageController(BaseController):
    READY = 'READY'
    BUSY = 'BUSY'
    SWEEP = 'SWEEP'
    STEPWISE = 'STEPWISE'

    @property
    def has_two_axes(self):
        return bool(self._has_two_axes)

    @property
    def is_busy(self):
        return self._device_status is not None and (self._worker_thread and self._worker_thread.is_alive())

    def __init__(self, serial_number=None, dlp_controller=None, monitor_interval=1.0):
        super(MockStageController, self).__init__(monitor_interval=monitor_interval)

        self._serial_number = serial_number
        self._device_info = None
        self._device = None
        self._device_status = MockStageController.READY  # set to None to make the mock unresponsive
        self._busy_status = None  # is SWEEP or STEPWISE or None
        self._dlp_controller = dlp_controller
        self._sweep_data = []
        self._worker_thread = None
        self._stop_worker_event = threading.Event()
        self._has_two_axes = False

    def get_angle(self):
        if self._device_status:
            return 1
        else:
            return 'N/A'

    def get_angle_y(self):
        if self._device_status:
            return 1
        else:
            return 'N/A'

    def get_info(self):
        return self._device_info

    def stop_motion(self):
        self._stop_worker_event.set()
        self._worker_thread = None

    def set_current_angle(self, angle):
        pass

    def move_to_angle(self, angle):
        if self.is_busy:
            raise StageBusyError()
        pass

    def move_to_angle_y(self, angle_y):
        if self.is_busy:
            raise StageBusyError()
        pass

    def jog(self, step):
        if self.is_busy:
            raise StageBusyError()
        pass

    def sweep(self, start_angle, end_angle, velocity, repetitions=1, time_interval=0.05):
        if self.is_busy:
            raise StageBusyError()
        self._stop_worker_event.clear()
        self._worker_thread = self._sweep_worker(start_angle, end_angle, velocity, repetitions, time_interval)

    def get_sweep_data(self):
        return self._sweep_data

    def stepwise(self, steps, delay_s):
        if self.is_busy:
            raise StageBusyError()
        self._stop_worker_event.clear()
        self._worker_thread = self._stepwise_worker(steps, delay_s)

    @run_in_thread
    def _sweep_worker(self, start_angle, end_angle, velocity, repetitions=1, time_interval=0.05):
        try:
            self._busy_status = MockStageController.SWEEP
            self._sweep_data = []
            while repetitions == -1 or repetitions > 0:
                if self._stop_worker_event.is_set():
                    return
                time.sleep(time_interval)
                if self._stop_worker_event.is_set():
                    return
                current_sweep = [(0, 0, 0)]
                if repetitions > 0:
                    self._sweep_data.append(current_sweep)
                    print(self._sweep_data)
                    repetitions -= 1
                if self._stop_worker_event.is_set():
                    return
        finally:
            self._busy_status = None

    @run_in_thread
    def _stepwise_worker(self, steps, delay_s):
        try:
            self._busy_status = MockStageController.STEPWISE
            for step in steps:
                if self._stop_worker_event.is_set():
                    break
                logger.info(step)
                time.sleep(delay_s)
        finally:
            self._busy_status = None

    def _connected_status(self):
        if self.is_busy:
            return self._busy_status or MockStageController.BUSY
        return MockStageController.READY

    def _disconnect(self):
        self.stop_motion()

    def _connect(self):
        self._device_info = {'serial_number': 0, 'model': 0, 'hardware_type': 0,
                             'firmware_version': 0, 'notes': 0, 'hardware_version': 0,
                             'modification_state': 0, 'no_of_channels':0}


    def _check_connection(self):
        last_angle = 1
        try:
            if last_angle != 'N/A':
                pass
        except:
            # persist the last known angle if the motor connection fails.
            if last_angle != 'N/A':
                raise
        return True


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    controller = MockStageController()
    controller.start()


    try:
        time.sleep(2)
        controller.stepwise((1, 2, 3), 1)
        while controller.is_busy:
            print(controller.status)
            time.sleep(0.5)
    finally:
        controller.stop()




import inspect
import sys
from phootonics_controller.utils.MR_E_2_Python_SDK.inflect import inflect
from ..tools.definitions import CommandID
from ..registers.ClassAbstracts import System

# # Currently not configured, must pull all classes from registers module
# def system_names():
#     all_systems = inspect.getmembers(sys.modules[__name__], inspect.isclass)
#     return [system[0] for system in all_systems if system[1]()._is_a_system]


def list_systems():
    """
    Inspect generic_registers for all imported and defined Systems.
    :return: list of System names
    """
    all_systems = inspect.getmembers(sys.modules['optoKummenberg.registers.generic_registers'], inspect.isclass)
    return [system[0] for system in all_systems if hasattr(system[1](0), '_is_a_system')]


def get_register(system_name: str, register_name, channel: int = 0):
    """
    Given a System name and Register name, will return the associated register_id integer
    :param system_name: str
    :param register_name: str
    :param channel: int
    :return: int register_id
    """
    registers = inspect.getmembers(sys.modules['optoKummenberg.registers.generic_registers'], inspect.isclass)
    register_system = dict(registers)[system_name](channel)
    register_list = inspect.getmembers(register_system)
    register_id = dict(register_list)[register_name]
    return register_id


def get_registers(system_name: str, channel: int = 0):
    """
    Returns list of tuples containing registers name and dictionary.
    :param system_name:
    :param channel:
    :return:
    """
    if inspect.isclass(system_name):
        if issubclass(system_name, System):
            system_object = system_name()
    else:
        all_systems = inspect.getmembers(sys.modules['optoKummenberg.registers.generic_registers'], inspect.isclass)
        system_object = dict(all_systems)[system_name](channel)

    attributes = inspect.getmembers(system_object, lambda a: not (inspect.isroutine(a)))
    register_list = [item for item in attributes if type(item[1]) is dict
                     and not (item[0].startswith('__') and item[0].endswith('__'))]

    return dict(register_list)


def process_registers(register_fields: dict, values):
    r"""
    Gives validity of registers field values to be set, and returns command and registers(s) IDs.

    Parameters
    ----------
    register_fields : dict
        The dictionary containing registers information.
    values
        The desired value(s) to set the registers.

    Returns
    -------
    command_id : int
        Command ID necessary to send single or multiple registers settings.
    register_id : int
        Register ID corresponding to given registers.

    Raises
    ------
    Returns None in case of parse_error (value out of range, or invalid input)

    See Also
    --------
    Register fields are returned by registers class attributes, see registers.

    """
    if isinstance(register_fields, list):
        if len(register_fields) > 8:
            print('Cannot Process more than 8 registers at once.')
            return None

        if values is None:
            pass
        else:
            if not isinstance(values, list):
                value_list = []
                for _ in range(len(register_fields)):
                    value_list.append(values)
                values = value_list
            if len(register_fields) != len(values):
                print("Number of Registers is Less Than Number of Values!")
                return None

        for i in range(len(register_fields)):
            if values is None:
                pass
            elif not is_valid_value(register_fields[i], values[i]):
                # invalid value sent!
                invalid_number = inflect.engine().ordinal(i+1)
                print('Invalid value for {} given registers'.format(invalid_number))
                valid_range = register_fields[i]['range']
                print('Given Value: {}. Valid Range: {}'.format(values, valid_range))
                return None
        register_id = [field['id'] for field in register_fields]
    else:
        if values is None:
            # get command
            pass
        elif not is_valid_value(register_fields, values):
            # invalid value sent!
            print('Invalid value for given registers')
            valid_range = register_fields['range']
            print('Given Value: {}. Valid Range: {}'.format(values, valid_range))
            return None
        register_id = register_fields['id']
    return register_id


def is_valid_channel(value):
    r"""
    Given a channel number value, determines if value is within accepted parameters.

    Parameters
    ----------
    value
        Must be between 0 and 8, can be int float or string

    """
    if type(value) in (int, float, str):
        if type(value) is int:
            return 0 <= value <= 8
        elif type(value) is float:
            return value.is_integer() and 0 <= value <= 8
        else:
            try:
                return 0 <= int(value) <= 8
            except:
                return False
    else:
        return False


def is_valid_value(register_dict, value):
    r"""
    Given a registers record and a value, determines if value is within accepted parameters.

    Parameters
    ----------
    register_dict : dict
        Register dictionary fields corresponding to desired registers.{'id': registers number, 'type': data type, ...}
    value : varies
        Value to set given registers.
    Raises
    ------
    none

    Notes
    -----
    Valid inputs follow the following logic table.

    +------------------------+-------------+-------------+------------+------------+
    | Input Value            | Register    | Register    | Register   | Register   |
    |                        | Type: int   | Type: float | Type: bool | Type: None |
    +========================+=============+=============+============+============+
    | int                    | check range | check range | is binary? | False      |
    +------------------------+-------------+-------------+------------+------------+
    | float                  | check range | check range | is binary? | False      |
    +------------------------+-------------+-------------+------------+------------+
    | bool                   | False       | False       | True       | False      |
    +------------------------+-------------+-------------+------------+------------+
    | None                   | False       | False       | False      | True       |
    +------------------------+-------------+-------------+------------+------------+
    | Other                  | False       | False       | False      | False      |
    +------------------------+-------------+-------------+------------+------------+

    Additional Notes
    ----------------
    Validation for requests to set values in the Signal Flow Manager are slightly different.
    The channel you are setting must match the channel in the flow manager.
    For example, it would not make sense to set the Input channel 0 to SPI channel 2, etc.
    """
    # TODO: Add list type for discrete ranges, specifically for SystemID range.
    try:
        value_type = type(value)
        register_type = register_dict['type']
        register_id = register_dict['id']
        register_range = register_dict['range']
        is_flow_manager = register_dict['unit'] == 'SystemID'
    except:
        return False

    if type(register_range) is dict:
        register_range = list(register_range.keys())

    if is_flow_manager:
        try:
            value = value.sys_id
        except:
            if value_type is not int:
                return False

        channel_id = (register_id & 0x00ff)//5
        if channel_id in [value % 8]:
            return True
        else:
            return False

    def range_check(val, valid_range):
        return valid_range[0] <= val <= valid_range[-1]

    if register_type is int:
        if value_type in (int, float):
            if value_type is int:
                if register_range is None:
                    return True
                else:
                    return range_check(value, register_range)
            if value_type is float:
                if value.is_integer():
                    if register_range is None:
                        return True
                    else:
                        return range_check(value, register_range)
                else:
                    return False
        elif value_type is bool:
            return value in (True, False)
        else:
            return False

    if register_type is float:
        if value_type in (int, float):
            if register_range is None:
                return True
            return range_check(value, register_range)
        else:
            return False

    if register_type is bool:
        return value in (True, False)

    if register_type is None:
        return value_type is None


# -*- coding: utf-8 -*-
"""
Created on September 01, 2015

Copyright Alpes Lasers SA, Neuchatel, Switzerland, 2015

@author: chiesa
"""

import argparse
import logging

logger = logging.getLogger(__name__)

_version = 'undefined'

try:
    from phootonics_controller import pkg, version

    _version = '{} {}'.format(pkg, version)
except Exception as e:
    logger.debug(e, exc_info=1)


baseparser = argparse.ArgumentParser(add_help=False)
baseparser.add_argument("--show-version", help="Show the project version",
                        action="version", version="%s %s" % (pkg, version))
alparser = argparse.ArgumentParser(parents=[baseparser])





# -*- coding: utf-8 -*-
"""
Created by Olga Reinauer

Copyright Alpes Lasers SA, Switzerland
"""
__author__ = 'olgare'
__copyright__ = "Copyright Alpes Lasers SA"

import logging
import threading
import time
from time import sleep

import numpy

from phootonics_controller.base_controllers.config import VCM_PORT, CAVITY1_INDEX, CAVITY2_INDEX, CAVITY3_INDEX, \
    CAVITY1_URL, CAVITY2_URL, CAVITY3_URL, ADC_CHANNEL, REF, configuration_dictionary, CAVITY4_URL, CAVITY4_INDEX, \
    EXT_VCM_PORT, get_scanning_wavelengths
from phootonics_controller.base_controllers.stage_mock import MockStageController
from phootonics_controller.base_controllers.xc_controller import ECController, MockECController
from phootonics_controller.base_controllers.stage_vcm import VCMirrorController

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def ignore_exception(f):
    def g(*args, **kargs):
        try:
            return f(*args, **kargs)
        except Exception as e:
            logger.exception(e)

    return g


class ADCDevice:

    def __init__(self):
        from phootonics_controller.utils.ADC_converter import ADS1263
        self._adc = ADS1263.ADS1263()

    def initialise(self):
        if self._adc.ADS1263_init() == -1:
            pass

    def read(self):
        try:
            adc_value = self._adc.ADS1263_GetAll()
            if adc_value[ADC_CHANNEL] >> 31 == 1:
                value = REF * 2 - adc_value[ADC_CHANNEL] * REF / 0x80000000
            else:
                value = adc_value[ADC_CHANNEL] * REF / 0x7fffffff  # 32bit
            return value
        except Exception as e:
            logger.error('could not read ADC value {}'.format(e))
            self._adc.ADS1263_Exit()


class MockADCDevice:

    def initialise(self):
        pass

    def read(self):
        return 0.2


class ScanCommand:

    def __init__(self, controller, wavelengths, positions):
        self._controller = controller
        self.wavelengths = wavelengths
        self.positions = positions
        self.results = {'active_cavity': [],
                        'x_set_point': [],
                        'x': [],
                        'y_set_point': [],
                        'y': [],
                        'wavelength_set_point': [],
                        'wavelength': [],
                        'detector_voltage': []}
        self.stopped = threading.Event()
        self.currentX = None
        self.currentY = None
        self.activeCavity = None
        self.currentWavelength = None
        self.detectorVoltage = None
        self.error = None
        self._testWithMocks = True
        self._completedWl = 0

    def execute(self):
        try:
            if not self._controller.testMode or self._testWithMocks:
                self._controller.stop_all_pulsing()
                self._controller.move_to_cavity_1()
                self._controller.get_active_cavity().activate_s2()
                for wavelength in self.wavelengths:
                    self._completedWl += 1
                    if self.stopped.is_set():
                        return
                    logger.info('going to wavelength {}'.format(wavelength))
                    cavity_index = self._controller.get_cavity_from_wavelength(wavelength)
                    if cavity_index is None:
                        logger.info('cannot reach wavelength {}: skip to next wavelength'.format(wavelength))
                        continue
                    logger.info('selecting cavity {}'.format(cavity_index))
                    self._controller.select_cavity(cavity_index)
                    cavity = self._controller.get_cavity_from_index(cavity_index)
                    cavity.move_to_wavelength(wavelength, timeout=2)
                    for posX, posY in self.positions:
                        if self.stopped.is_set():
                            return
                        self._controller.extVcm.move_to_angle_y(posY)
                        self._controller.extVcm.move_to_angle(posX)
                        currentX = self._controller.extVcm.get_angle()
                        currentY = self._controller.extVcm.get_angle_y()
                        activeCavity = self._controller.get_active_cavity_index()

                        detectorVoltage = self._controller.adc.read()
                        self.results['active_cavity'].append(activeCavity)
                        self.results['x_set_point'].append(posX)
                        self.results['x'].append(currentX)
                        self.results['y_set_point'].append(posY)
                        self.results['y'].append(currentY)
                        self.results['wavelength_set_point'].append(wavelength)
                        self.results['wavelength'].append(cavity.get_wavelength())
                        self.results['detector_voltage'].append(detectorVoltage)

                        self.activeCavity = activeCavity
                        self.currentX = currentX
                        self.currentY = currentY
                        self.detectorVoltage = detectorVoltage


            else:
                if not self.stopped.wait(2.0):
                    self.results = {'active_cavity': [1, 1, 2, 2, 3, 3, 4, 4],
                                    'x_set_point': [0, 1, 0, 1, 0, 1, 0, 1],
                                    'x': [0.2, 1.2, 0.2, 1.2, 0.2, 1.2, 0.2, 1.2],
                                    'y_set_point': [0, 1, 0, 1, 0, 1, 0, 1],
                                    'y': [0.23, 1.23, 0.23, 1.23, 0.23, 1.23, 0.23, 1.23],
                                    'wavelength_set_point': [1300, 1301, 1302, 1303, 1304, 1305, 1306, 1307],
                                    'wavelength': [1300.2, 1301.1, 1302.4, 1303.3, 1304.2, 1305.6, 1306.2, 1307.1],
                                    'detector_voltage': [0.2, 0.5, 0.3, 0.11, 0.2, 0.5, 0.3, 0.11]}

        except Exception as e:
            self.error = str(e)
            logger.exception(e)
        finally:
            self._controller.stop_all_pulsing()

    def stop(self):
        self.stopped.set()

    def to_dict(self):
        return {'runtime': {'active_cavity': self.activeCavity,
                            'current_X': self.currentX,
                            'current_y': self.currentY,
                            'error': self.error,
                            'detector_voltage': self.detectorVoltage,
                            'completed': self._completedWl/len(self.wavelengths)},
                'results': self.results}


class FastScanCommand:

    def __init__(self, controller, wavelengths, amplitude_x, amplitude_y):
        self._controller = controller
        self.wavelengths = wavelengths
        self.amplitude_x = amplitude_x
        self.amplitude_y = amplitude_y
        self.results = {'active_cavity': [],
                        'x_set_point': [],
                        'x': [],
                        'y_set_point': [],
                        'y': [],
                        'wavelength_set_point': [],
                        'wavelength': [],
                        'detector_voltage': []}
        self.stopped = threading.Event()
        self._acqLockDetector = threading.RLock()
        self._acqLockVcm = threading.RLock()
        self.currentX = None
        self.currentY = None
        self.activeCavity = None
        self.currentWavelength = None
        self.detectorVoltage = None
        self.error = None
        self._testWithMocks = True
        self._completedWl = 0
        self._detectorAcquisitionThread = None
        self._detectorAcquisitions = []
        self._vcmAcquisitions = []

    def start_detector_acquisition(self):
        self._detectorAcquisitionThread = threading.Thread(target=self._run_acquisition)
        self._detectorAcquisitionThread.start()

    def _run_acquisition(self):
        while True:
            if self.stopped.is_set():
                return
            with self._acqLockDetector:
                self._detectorAcquisitions.append((time.time(), self._controller.adc.read()))
            if self.stopped.wait(0.001):
                return

    def record_callback(self, x, y, t):
        with self._acqLockVcm:
            self._vcmAcquisitions.append((t, x, y))

    def clear_registers(self):
        with self._acqLockDetector:
            self._detectorAcquisitions = []
        with self._acqLockVcm:
            self._vcmAcquisitions = []

    def execute(self):
        try:
            if not self._controller.testMode or self._testWithMocks:
                self._controller.stop_all_pulsing()
                self._controller.move_to_cavity_1()
                self._controller.get_active_cavity().activate_s2()
                self.start_detector_acquisition()
                self._controller.extVcm._device.start_x_signal_generator(self.amplitude_x, 25, 0, 0)
                self._controller.extVcm._device.start_y_signal_generator(self.amplitude_y, 1, 0, 0)
                self._controller.extVcm._device.register_record_callback(self.record_callback)
                for wavelength in self.wavelengths:
                    self._completedWl += 1
                    if self.stopped.is_set():
                        return
                    logger.info('going to wavelength {}'.format(wavelength))
                    cavity_index = self._controller.get_cavity_from_wavelength(wavelength)
                    if cavity_index is None:
                        logger.info('cannot reach wavelength {}: skip to next wavelength'.format(wavelength))
                        continue
                    logger.info('selecting cavity {}'.format(cavity_index))
                    self._controller.select_cavity(cavity_index)
                    cavity = self._controller.get_cavity_from_index(cavity_index)
                    cavity.move_to_wavelength(wavelength)

                    self.clear_registers()

                    if self.stopped.wait(1):
                        return

                    detector = numpy.array(self._detectorAcquisitions)
                    vcm = numpy.array(self._vcmAcquisitions)

                    detectorTimes = detector[:, 0]
                    detectorValues = detector[:, 1]
                    vcmTimes = vcm[:, 0]
                    vcmX = vcm[:, 1]
                    vcmY = vcm[:, 2]

                    self.clear_registers()

                    interpX = numpy.interp(detectorTimes, vcmTimes, vcmX)
                    interpY = numpy.interp(detectorTimes, vcmTimes, vcmY)

                    activeCavity = self._controller.get_active_cavity_index()
                    self.results['active_cavity'].append(activeCavity)

                    self.results['x'].append(interpX.tolist())
                    self.results['y'].append(interpY.tolist())
                    self.results['wavelength_set_point'].append(wavelength)
                    self.results['wavelength'].append(cavity.get_wavelength())
                    self.results['detector_voltage'].append(detectorValues.tolist())

                    self.activeCavity = activeCavity
                    self.currentX = None
                    self.currentY = None
                    self.detectorVoltage = None


            else:
                if not self.stopped.wait(2.0):
                    self.results = {'active_cavity': [1, 1, 2, 2, 3, 3, 4, 4],
                                    'x_set_point': [0, 1, 0, 1, 0, 1, 0, 1],
                                    'x': [0.2, 1.2, 0.2, 1.2, 0.2, 1.2, 0.2, 1.2],
                                    'y_set_point': [0, 1, 0, 1, 0, 1, 0, 1],
                                    'y': [0.23, 1.23, 0.23, 1.23, 0.23, 1.23, 0.23, 1.23],
                                    'wavelength_set_point': [1300, 1301, 1302, 1303, 1304, 1305, 1306, 1307],
                                    'wavelength': [1300.2, 1301.1, 1302.4, 1303.3, 1304.2, 1305.6, 1306.2, 1307.1],
                                    'detector_voltage': [0.2, 0.5, 0.3, 0.11, 0.2, 0.5, 0.3, 0.11]}

        except Exception as e:
            self.error = str(e)
            logger.exception(e)
        finally:
            self.stop()
            self._controller.stop_all_pulsing()
            self._controller.extVcm._device.reset_static_input()
            self._controller.extVcm._device.unregister_record_callback(self.record_callback)

    def stop(self):
        self.stopped.set()

    def to_dict(self):
        return {'runtime': {'active_cavity': self.activeCavity,
                            'current_X': self.currentX,
                            'current_y': self.currentY,
                            'error': self.error,
                            'detector_voltage': self.detectorVoltage,
                            'completed': self._completedWl/len(self.wavelengths)},
                'results': self.results}




class AlreadyScanningException(Exception):
    pass


class ControllerNotReady(Exception):
    pass


class MainController:
    RESPONDING = 'RESPONDING'
    NOT_RESPONDING = 'NOT_RESPONDING'
    READY = 'READY'
    NOT_READY = 'NOT_READY'

    def __init__(self, loop_delay=0.005, test_mode=False):
        self._scanCommand = None
        self._scanCommandThread = None
        self.testMode = test_mode
        if not self.testMode:
            self.vcm = VCMirrorController(VCM_PORT)
            self.extVcm = VCMirrorController(EXT_VCM_PORT)
            self.adc = ADCDevice()
            self.ec1 = ECController(ec_url=CAVITY1_URL, monitor_interval=0.3, index=CAVITY1_INDEX)
            self.ec2 = ECController(ec_url=CAVITY2_URL, monitor_interval=0.3, index=CAVITY2_INDEX)
            self.ec3 = ECController(ec_url=CAVITY3_URL, monitor_interval=0.3, index=CAVITY3_INDEX)
            self.ec4 = ECController(ec_url=CAVITY4_URL, monitor_interval=0.3, index=CAVITY4_INDEX)
        else:
            self.vcm = MockStageController()
            self.extVcm = MockStageController()
            self.adc = MockADCDevice()
            self.ec1 = MockECController(monitor_interval=0.3, index=CAVITY1_INDEX)
            self.ec2 = MockECController(monitor_interval=0.3, index=CAVITY2_INDEX)
            self.ec3 = MockECController(monitor_interval=0.3, index=CAVITY3_INDEX)
            self.ec4 = MockECController(monitor_interval=0.3, index=CAVITY4_INDEX)
        self._active_cavity = None
        self._monitoring_data = {}
        self._recorded_data = {}
        self._monitoring_thread = None
        self._stop_monitoring_event = threading.Event()
        self._condition_thread = None
        self._stop_condition_event = threading.Event()
        self._monitoring_loop_delay = loop_delay
        self._condition_loop_delay = 0.5
        self._system_connection_status = None
        self._system_ready_status = None
        self._system_active_cavity = None
        self._ec1_connection_status = None
        self._ec1_ready_status = None
        self._ec2_connection_status = None
        self._ec2_ready_status = None
        self._ec3_connection_status = None
        self._ec3_ready_status = None
        self._ec4_connection_status = None
        self._ec4_ready_status = None
        self._vcm_status = None
        self._extVcm_status = None

    def get_cavity_from_wavelength(self, wavelength):
        for key, item in configuration_dictionary.items():
            if item['wavelength'][0] <= wavelength <= item['wavelength'][1]:
                return key
        return None

    def get_results(self):
        if self._scanCommand:
            return self._scanCommand.to_dict()

    def _check_ready(self):
        status = self.get_connection_status()
        logger.info('Phootonics controller status: {}'.format(status))
        if status != MainController.RESPONDING:
            raise ControllerNotReady
        if not self.is_ready_to_action() == 'READY':
            raise ControllerNotReady

    def start_scan(self, wl_step=None, positions=None):
        self._check_ready()
        if self._scanCommandThread:
            if self._scanCommandThread.is_alive():
                raise AlreadyScanningException
            self._scanCommandThread = None
        self._scanCommand = ScanCommand(controller=self,
                                        wavelengths=get_scanning_wavelengths(wl_step),
                                        positions=positions)
        self._scanCommandThread = threading.Thread(target=self._scanCommand.execute)
        self._scanCommandThread.start()

    def start_fast_scan(self, wl_step=None, amplitude_x=None, amplitude_y=None):
        self._check_ready()
        if self._scanCommandThread:
            if self._scanCommandThread.is_alive():
                raise AlreadyScanningException
            self._scanCommandThread = None
        self._scanCommand = FastScanCommand(controller=self,
                                            wavelengths=get_scanning_wavelengths(wl_step),
                                            amplitude_x=amplitude_x,
                                            amplitude_y=amplitude_y)
        self._scanCommandThread = threading.Thread(target=self._scanCommand.execute)
        self._scanCommandThread.start()

    def is_scan_command_running(self):
        if not self._scanCommandThread:
            return False
        return self._scanCommandThread.is_alive()

    def stop_scan(self):
        if self._scanCommand:
            self._scanCommand.stop()

    def start(self):
        try:
            logging.info('Initialization started, can take a few seconds')
            self.ec1.start()
            logging.info('Cavity 1 initialized')
            self.ec2.start()
            logging.info('Cavity 2 initialized')
            self.ec3.start()
            logging.info('Cavity 3 initialized')
            self.ec4.start()
            logging.info('Cavity 4 initialized')
            self.vcm.start()
            self.extVcm.start()
            logging.info('cavities initialized')
            self.adc.initialise()
            logging.info('system initialized')
            self.monitor()
        except Exception as e:
            logging.error('Could not start external cavity controllers {}'.format(e))
            self.shutdown_all_systems()

    def get_connection_status(self):
        if self._ec1_connection_status == 'RESPONDING' and self._ec2_connection_status == 'RESPONDING' and \
                self._ec3_connection_status == 'RESPONDING' and \
                self._vcm_status not in [None, 'DISCONNECTED', 'CONNECTION_ERROR'] and \
                self._extVcm_status not in [None, 'DISCONNECTED', 'CONNECTION_ERROR'] and \
                self._ec4_connection_status == 'RESPONDING':
            return MainController.RESPONDING
        else:
            return MainController.NOT_RESPONDING

    def is_ready_to_action(self):
        if self._ec1_ready_status == 'READY' and self._ec2_ready_status == 'READY' and \
                self._ec3_ready_status == 'READY' and \
                self._vcm_status == 'READY' and \
                self._extVcm_status == 'READY' and \
                self._ec4_ready_status == 'READY':
            return MainController.READY
        else:
            return MainController.NOT_READY

    def stop_all_pulsing(self):
        self.ec1.stop_pulsing()
        self.ec2.stop_pulsing()
        self.ec3.stop_pulsing()
        self.ec4.stop_pulsing()

    def shutdown_all_systems(self, reset=False):
        try:
            self.stop_monitoring()
            if reset:
                self.ec1.shutdown()
                self.ec2.shutdown()
                self.ec3.shutdown()
                self.ec4.shutdown()
            else:
                self.ec1.stop_pulsing()
                self.ec2.stop_pulsing()
                self.ec3.stop_pulsing()
                self.ec4.stop_pulsing()
            self.vcm.stop()
            self.extVcm.stop()
        except Exception as e:
            logging.error('error on shutdown {}'.format(e))

    def get_active_cavity_index(self):
        return self._active_cavity

    def get_active_cavity(self):
        return self.get_cavity_from_index(self._active_cavity)

    def get_cavity_from_index(self, cavity_index):
        if cavity_index == CAVITY1_INDEX:
            return self.ec1
        elif cavity_index == CAVITY2_INDEX:
            return self.ec2
        elif cavity_index == CAVITY3_INDEX:
            return self.ec3
        elif cavity_index == CAVITY4_INDEX:
            return self.ec4

    def stop_monitoring(self):
        self._stop_monitoring_event.set()
        self._monitoring_thread.join(1.0)

    def monitor(self):
        self._stop_monitoring_event.clear()
        self._monitoring_thread = threading.Thread(target=self._monitoring_worker)
        self._monitoring_thread.start()

    def get_monitoring_data(self):
        return self._monitoring_data

    def _monitoring_worker(self):
        first_time = time.time()
        ec1_monitoring_trace = None
        ec2_monitoring_trace = None
        ec3_monitoring_trace = None
        ec4_monitoring_trace = None
        t = None
        while self._stop_monitoring_event.wait(self._monitoring_loop_delay) is False:
            if not self._stop_monitoring_event.is_set():
                device_time = time.time()
                t = int(device_time - first_time)
                self._system_connection_status = ignore_exception(self.get_connection_status)()
                self._system_ready_status = ignore_exception(self.is_ready_to_action)()
                self._system_active_cavity = ignore_exception(self.get_active_cavity_index)()
                self._ec1_connection_status = self.ec1.status
                self._ec1_ready_status = ignore_exception(self.ec1.xc_ready_to_action)()
                self._ec2_connection_status = self.ec2.status
                self._ec2_ready_status = ignore_exception(self.ec2.xc_ready_to_action)()
                self._ec3_connection_status = self.ec3.status
                self._ec3_ready_status = ignore_exception(self.ec3.xc_ready_to_action)()
                self._ec4_connection_status = self.ec4.status
                self._ec4_ready_status = ignore_exception(self.ec4.xc_ready_to_action)()
                self._vcm_status = self.vcm.status
                self._extVcm_status = self.extVcm.status
                ec1_monitoring_trace = ignore_exception(self.ec1.system_info)()['response']['data']
                ec2_monitoring_trace = ignore_exception(self.ec2.system_info)()['response']['data']
                ec3_monitoring_trace = ignore_exception(self.ec3.system_info)()['response']['data']
                ec4_monitoring_trace = ignore_exception(self.ec4.system_info)()['response']['data']
            self._monitoring_data = {'t': t,
                                     'active_cavity': self._system_active_cavity,
                                     'system_ready_status': self._system_ready_status,
                                     'system_connection_status': self._system_connection_status,
                                     'ec1_connection': self._ec1_connection_status,
                                     'ec2_connection': self._ec2_connection_status,
                                     'ec3_connection': self._ec3_connection_status,
                                     'ec4_connection': self._ec4_connection_status,
                                     'ec1_ready': self._ec1_ready_status,
                                     'ec2_ready': self._ec2_ready_status,
                                     'ec3_ready': self._ec3_ready_status,
                                     'ec4_ready': self._ec4_ready_status,
                                     'ec1_trace': ec1_monitoring_trace,
                                     'ec2_trace': ec2_monitoring_trace,
                                     'ec3_trace': ec3_monitoring_trace,
                                     'ec4_trace': ec4_monitoring_trace,
                                     'vcm_status': self._vcm_status,
                                     'extVcm_status': self._extVcm_status
                                     }

    def select_cavity(self, cavity_index):
        if self._system_connection_status == MainController.RESPONDING:
            if self._active_cavity != cavity_index:
                logger.info('switching to cavity {}'.format(cavity_index))
                self._switch_and_activate_cavity(cavity_index, configuration_dictionary[cavity_index]['x_position'],
                                                 configuration_dictionary[cavity_index]['y_position'])
        else:
            raise Exception('Controller not responding')

    def _switch_and_activate_cavity(self, cavity_index, cavity_position_x, cavity_position_y):
        try:
            cavity = self.get_cavity_from_index(cavity_index)
            logger.info('active cavity stops pulsing {}'.format(self.get_active_cavity_index()))
            self.get_active_cavity().stop_pulsing()
            time.sleep(1)
            self._active_cavity = None
            logger.info('moving to position {} {}'.format(cavity_position_x, cavity_position_y))
            logger.info('VCM status {}'.format(self.vcm.status))
            self.vcm.move_to_angle(cavity_position_x)
            self.vcm.move_to_angle_y(cavity_position_y)
            time.sleep(1)
            cavity.activate_s2()
            time.sleep(1)
            self._active_cavity = cavity_index
        except Exception as e:
            logging.error('Could not move to cavity {} at position {} {}, error {}'.format(cavity_index,
                                                                                           cavity_position_x,
                                                                                           cavity_position_y, e))

    def move_to_cavity_1(self):
        try:
            self.vcm.move_to_angle(configuration_dictionary[CAVITY1_INDEX]['x_position'])
            self.vcm.move_to_angle_y(configuration_dictionary[CAVITY1_INDEX]['y_position'])
            self._active_cavity = CAVITY1_INDEX
        except Exception as e:
            logging.error('Could not move to cavity 1'.format(e))

    def get_adc_voltage(self):
        return self.adc.read()



import struct
import spidev
import time


class SPI:
    def __init__(self, bus: int = 0, device: int = 0):
        self._bus = bus
        self._device = device
        self._spi_comm = spidev.SpiDev()
        self.connect()
        self._spi_comm.max_speed_hz = 125000000//16
        self._spi_comm.mode = 0b01
        self._spi_comm.lsbfirst = False
        self.x_readback = 0
        self.y_readback = 0
        self.reg_value = 0
        self.status = None

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()

    def set_values(self, system1, register1, system2, register2, data1, data2, data_type='>f'):
        ans = self.send(system1, register1, system2, register2, data1, data2, data_type)
        return [hex(i) for i in ans]

    def get_values(self, system, register, unit_type='>f'):
        resp = self._spi_comm.xfer([0, 0, system, register, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
        return self.receive_r(resp, unit_type)

    def send(self,  system1, register1, system2, register2, data1, data2, data_type='>f'):
        byte_string = struct.pack('>B', 0) + \
                      struct.pack('>B', 1) + \
                      struct.pack('>B', system1) + \
                      struct.pack('>B', register1) + \
                      struct.pack('>B', system2) + \
                      struct.pack('>B', register2) + \
                      struct.pack(data_type, data1) + \
                      struct.pack(data_type, data2)
        print(byte_string.hex())
        return self._spi_comm.xfer(byte_string)

    def receive_r(self, response, unit_type):
        self.reg_value = struct.unpack(unit_type, bytes(response[2:6]))[0]
        self.x_readback = struct.unpack('>f', bytes(response[6:10]))[0]
        self.y_readback = struct.unpack('>f', bytes(response[10:14]))[0]

        return self.reg_value, self.x_readback, self.y_readback

    def connect(self):
        self._spi_comm.open(self._bus, self._device)

    def disconnect(self):
        self._spi_comm.close()



import serial
from .registers import generic_registers as registers
from .tools.definitions import UnitType, CommandID, FRAME_BOUNDARY, SimpleResponses, CMD_DELAY
from .commands import Command
from .tools.command_tools import issue_command
import time
from warnings import warn
from sys import exit

from logging import getLogger

logger = getLogger(__name__)


class Connection:
    r"""
    Connection object from which to connect/disconnect and send/receive data.

    For Serial Communication, See Also: PySerial API Documentation
        https://pyserial.readthedocs.io/en/latest/pyserial_api.html
    or use help(serial).

    Parameters
    ----------
    port : str or None
        The port is immediately opened on object creation, when a port is given.
        It is not opened when port is None and a successive call to open() is required.
        port is a device name: depending on operating system. e.g. /dev/ttyUSB0 on GNU/Linux or COM3 on Windows.
    baudrate : int
        The parameter baudrate can be one of the standard values:
        50, 75, 110, 134, 150, 200, 300, 600, 1200, 1800, 2400, 4800, 9600, 19200, 38400, 57600, 115200.
        These are well supported on all platforms.
        Standard values above 115200, such as:
        230400, 460800, 500000, 576000, 921600, 1000000, 1152000, 1500000, 2000000, 2500000, 3000000, 3500000, 4000000
        also work on many platforms and devices.
    comm_lock : bool = True
        Communication is disabled if this parameter is True, which is default. This prevents sending and receiving over
        the channel.
    timeout : float
        timeout = None: wait forever / until requested number of bytes are received
        timeout = 0: non-blocking mode, return immediately in any case, returning zero or more,
            up to the requested number of bytes
        timeout = x: set timeout to x seconds (float allowed) returns immediately when the requested
            number of bytes are available, otherwise wait until the timeout expires and return all
            bytes that were received until then.
    verbose : bool
        Verbose serial output for debugging/etc. If true, output generated even without parse_error/failures.

    Raises
    ------
    SerialException
        Unable to establish connection over given port.

    Notes
    -----
    To prevent communication issues, make sure that the comm_lock is False for only one device at a time.

    Examples
    --------
    Instantiate a connection on COM3, with baud rate 9600

    >>> con = Connection('COM3', 9600, verbose=True)
    Establishing Connection via port [COM3]...
    Connected.

    """

    def __init__(self, port: str or None, baudrate: int, inter_byte_timeout: float,
                 comm_lock: bool, timeout: float, verbose: bool):
        self._port               = port
        self._baudrate           = baudrate
        self._inter_byte_timeout = inter_byte_timeout
        self._comm_lock          = comm_lock
        self._timeout            = timeout
        self._serial_conn        = None
        self._settings           = None
        self.verbose             = verbose
        self.last_sent           = None

        if port is not None:
            # make connection to input
            self.connect()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disconnect()

    @staticmethod
    def _read(serial_connection, length: int = None, terminator: bytes = None):
        if terminator is None:
            if length is None:
                return serial_connection.read(serial_connection.inWaiting())
            else:
                return serial_connection.read(size=length)
        else:
            return serial_connection.read_until(terminator)

    def send(self, data: bytes):
        r"""
        Send bytes array over connection.

        Parameters
        ----------
        data : bytes
            Bytes object to send.

        Returns
        -------
        none

        Raises
        ------
        none

        Notes
        -----
        Bytes must be properly encoded, see example.

        Examples
        --------
        First line: Set _board to pro mode.
        Second Line: Set value (0x10) of USB/UART (0x50) OF (0x00) registers to 0.
        Third Line: Same as above, without using encoder.

        >>> con = Connection('COM12', 9600, verbose=True)
        Establishing Connection via port [COM3]...
        Connected.
        >>> con._comm_lock = False
        >>> con.send('gopro\r\n'.encode())
        DEBUG: Sending 'b'gopro\r\n''
        >>> con.send(con.encode(0x10,0x5000,0))
        DEBUG: Sending 'b'~\x00\x10\x06P\x00\x00\x00\x00\x00\x00\x00~''
        >>> con.send(b'~\x00\x10\x06P\x00\x00\x00\x00\x00\x00\x00~')
        DEBUG: Sending 'b'~\x00\x10\x06P\x00\x00\x00\x00\x00\x00\x00~''

        """
        if self._port is not None:
            try:
                self._serial_conn.write(data)
                if self.verbose:
                    if data[-1] == 0x0a:
                        logger.debug('Sending: {}'.format(data))
                    else:
                        logger.debug('Sending: {}'.format(data.hex()))
            except AttributeError:
                logger.debug('WRITE ERROR: No Connection')
        elif self.verbose:
            logger.debug('DEBUG: Sending \'{}\''.format(data))
        self.last_sent = data

    def receive(self, length: int = None, terminator: bytes = None):
        r"""
        Receive bytes array over connection.

        Parameters
        ----------
        length : int, optional
            Return given number of bytes.
        terminator : bytes, optional
            The communication will proceed until the terminator is received.
            The remaining data will still exist in the firmware buffer.

        Returns
        -------
        bytes array
            Encoded byte array returned.

        Raises
        ------
        none

        Examples
        --------
        Assuming the 'gopro' command was just sent, a response will be transmitted by the MRE2 firmware.

        >>> con.receive()
        Received: b'OK\r\n'

        """
        if self._port is not None:
            data = self._read(self._serial_conn, length, terminator)
            if terminator == bytes([FRAME_BOUNDARY]):
                data += self._read(self._serial_conn, length, terminator)
                while data == terminator:
                    data += self._read(self._serial_conn, length, terminator)
            if self.verbose:
                if 0x0a in data:
                    if data[-1] == 0x0a:
                        logger.debug('Received: {}'.format(data))
                    else:
                        logger.debug('Received: {}'.format(data.hex()))
                else:
                    logger.debug('Received: {}'.format(data.hex()))

        else:
            data = b'OK\r\n'
            if self.verbose:
                logger.debug('DEBUG: Received Simulated Response {}'.format(data))

        return data

    def connect(self):
        r"""
        Establish serial connection over port. Will reattempt up to 3 times.

        Parameters
        ----------
        none

        Returns
        -------
        none

        Raises
        ------
        SerialException
        """
        if self.verbose:
            logger.debug('Establishing Connection via port [{}]...'.format(self._port))
        if self._port is not None:
            attempts = 0
            while attempts < 3:
                try:
                    # connect
                    attempts += 1
                    self._serial_conn = serial.Serial(port=self._port, baudrate=self._baudrate, timeout=self._timeout,
                                                      interCharTimeout=self._inter_byte_timeout)
                    self._settings = self._serial_conn.getSettingsDict()
                    if self.verbose:
                        logger.debug('Connected.')
                    break
                except serial.serialutil.SerialException as e:
                    logger.exception(e)
                    err_args = serial.serialutil.SerialException.args
                    if hasattr(self, '_serial_conn'):
                        del self._serial_conn
                    logger.debug('Unable to Connect.')
                    logger.debug('Retrying Connection...')
                    pass
            if attempts == 3:
                logger.debug('Exiting.')
                raise Exception('cannot connect to VCM after 3 attempts')
        elif self.verbose:
            logger.debug('DEBUG MODE: Connection Simulated.')

    def disconnect(self):
        r"""
        Terminates the connection with a given _board.

        Returns
        -------
        none

        Raises
        ------
        none

        Examples
        --------
        >>> con.disconnect()
        Flushing input-output buffers...
        Closing Connection...
        Disconnected.

        """
        if self._port is not None:
            if self.verbose:
                logger.debug('Flushing input-output buffers...')
            self._serial_conn.flush()
            if self.verbose:
                logger.debug('Closing Connection...')
            self._serial_conn.close()
        if self.verbose:
            logger.debug('Disconnected.')


class Board(Command):
    r"""
    Board object from which to issue MRE2 commands over connection.

    Parameters
    ----------
    port, baudrate, comm_lock, timeout, verbose
        Extended from Connection class
    simple: bool, optional
        Determines whether or not Board will be set in simple-mode after initialization.
    board_reset: bool, optional
        Determines whether or not to reset the Board during initialization.
    number_of_channels: int
        How many channel will be instantiated during initalization.

    Raises
    ------
    none

    Notes
    -----
    Board extends Connection class. Refer to Connection for parameters, etc.

    Examples
    --------
    Instantiate a simple-mode Board with connection over COM12 port, with baud rate 9600. Then switch into pro-mode.
    This will not flush the firmware registers, so it is assumed to already be reset.

    >>> board = Board(port='COM12', baudrate=9600, verbose=True, simple=True)
    Establishing Connection via port [COM12]...
    Connected.
    Initializing Board in Simple Mode...
    Sending: b'~\x00\x06\x02\x00\x00\x00\x00~'
    Received: b''
    Verifying Simple via Handshake...
    Sending: b'Start\r\n'
    Received: b'NO\r\n'
    Handshake NOT Successful: UNKNOWN_ERROR_CODE.
    Sending: b'Start\r\n'
    Received: b'OK\r\n'
    Handshake Successful.
    Board Initialized in Simple Mode.
    Initialization Finalized. Board is Ready.
    >>> board.go_pro()
    Sending: b'GoPro\r\n'
    Received: b'OK\r\n'
    Go Pro Successful.

    """
    def __init__(self, port: str or None = None, baudrate: int = 115200, comm_lock: bool = True,
                 timeout: float = None, verbose: bool = False, simple: bool = False,
                 board_reset: bool = False, inter_byte_timeout: float = 0.2):
        self.verbose = verbose
        self.simulated = port is None
        if self.simulated:
            warn("This is a simulated connection: Serial Port = None. "
                 "If this is unintended, try explicitly stating the port", RuntimeWarning)
        # default board-level systems
        self.Status = registers.Status(board=self)
        self.EEPROM = registers.BoardEEPROM(board=self)
        self.SnapshotManager = registers.SnapshotManager(board=self)

        # default proxy board
        self.ProxyBoard = ProxyBoard(board=self)

        # default channels
        self.channel = [self.ProxyBoard.Channel_0]

        # default connection and Kummenberg commands
        self.Connection = Connection(port, baudrate, inter_byte_timeout, comm_lock, timeout, verbose)
        self.Commands = Command()
        Command.__init__(self)
        self._simple = self.IsSimpleMode()
        self._set_internal_mode(self._simple, simple)

    def ResetChannel(self, channel_to_reset: int):
        r"""
        Reset given channel to static input in open-loop mode, with input value 0. Reset Gain/Offset.

        Returns
        -------
        none

        Raises
        ------
        none

        Examples
        --------
        >>> board.reset_channel(0)
        Flushing input-output buffers...
        Closing Connection...
        Disconnected.

        """
        if len(self.channel) > channel_to_reset:
            if self.verbose:
                logger.debug('Resetting Channel {} to Static Input with 0 Amps...'.format(channel_to_reset))
            self.channel[channel_to_reset].SetControlMode(UnitType.CURRENT)
            self.channel[channel_to_reset].StaticInput.SetCurrent(0.0)
            self.channel[channel_to_reset].StaticInput.SetAsInput()
            self.channel[channel_to_reset].InputConditioning.SetGain(1.0)
            self.channel[channel_to_reset].InputConditioning.SetOffset(0.0)
            if self.verbose:
                logger.debug('Channel {} Reset to Default State.'.format(channel_to_reset))
        else:
            logger.debug('Channel {} is not a Valid Channel Number.'.format(channel_to_reset))

    def IsSimpleMode(self):
        r"""
        Will determine if board is in simple-mode.

        Returns
        -------
        is_simple_mode: bool

        Raises
        ------
        none

        Examples
        --------
        >>> board.IsSimpleMode()

        """
        self.Connection._comm_lock = False
        started_correctly = False
        num_attempts = 0
        ack = b''

        while not started_correctly and num_attempts < 3:
            self.Connection.send(b'Start\r\n')
            time.sleep(CMD_DELAY)
            ack = self.Connection.receive()
            if ack != b'':
                started_correctly = True
            else:
                num_attempts += 1
                time.sleep(CMD_DELAY)

        if ack in SimpleResponses.values():
            is_simple_mode = True
        # we enter here if we received pro garbage OR
        # if we havent received anz response in the 3 Start attempts
        else:
            status_cmd = bytes([FRAME_BOUNDARY, 0, CommandID.GET_STATUS, 2, 0, 0, 0, 0, FRAME_BOUNDARY])
            issue_command(self, status_cmd)
            is_simple_mode = False
        self.Connection._comm_lock = True
        return is_simple_mode

    def _set_internal_mode(self, current_mode_is_simple, simple_mode_desired):
        if simple_mode_desired:
            if current_mode_is_simple:
                self._simple = True
            elif not current_mode_is_simple:
                if self.verbose:
                    logger.debug('Initializing Board in Simple Mode...')
                self.go_simple()
                if self.verbose:
                    logger.debug('Board Initialized in Simple Mode.')
        elif not simple_mode_desired:
            if current_mode_is_simple:
                if self.verbose:
                    logger.debug('Initializing Board in Pro Mode...')
                self.go_pro()
                if self.verbose:
                    logger.debug('Board Initialized in Pro Mode.')
            elif not current_mode_is_simple:
                self._simple = False
        return


class ProxyBoard:
    r"""
    Generic Proxy Board object with a single channel.

    Parameters
    ----------
    board
        connections.Board object

    Raises
    ------
    none

    Notes
    -----
    Board extends Connection class. Refer to Connection for parameters, etc.

    """
    def __init__(self, board):
        self._board = board
        self.Channel_0 = Channel(board)

    def GetConnectedStatus(self):
        r"""
        Determines whether proxy board is connected and responding.

        Parameters
        ----------
        none

        Raises
        ------
        none

        Returns
        -----
        bool

        """
        return not self._board.Status.GetErrorFlagRegister()


class Channel:
    r"""
    Generic Channel object with default systems and commands. Such as StaticInput, SetControlMode, etc

    Parameters
    ----------
    board
        connections.Board object
    channel: int
        channel identifier, required for system and register identification

    Raises
    ------
    none
    """
    def __init__(self, board, channel_number: int = 0):
        self._board = board
        self._channel = channel_number
        self._control_mode = UnitType.CURRENT

        # default meta systems
        self.Manager           = registers.Manager(self._channel, self._board)

        # default stages
        self.StaticInput       = registers.StaticInput(self._channel, self._board)
        self.InputConditioning = registers.InputConditioning(self._channel, self._board)
        self.FeedThrough       = registers.CurrentFeedThrough(self._channel, self._board)
        self.LinearOutput      = registers.LinearOutput(self._channel, self._board)

    def GetControlMode(self):
        r"""
        Retrieves board control mode for this channel.

        Parameters
        ----------
        none

        Raises
        ------
        none

        Returns
        -----
        UnitType
            Control mode options set by register_tables.UnitType

        """
        response = self.Manager.get_register('control')

        if response[0] == 0xB0 | self._channel:  # Feed Through Control Mode
            self._control_mode = UnitType.CURRENT
            return self._control_mode

        if response[0] == 0xB8 | self._channel:  # OF PID Control Mode
            self._control_mode = UnitType.OF
            return self._control_mode

        if response[0] == 0xC0 | self._channel:  # XY PID Control Mode
            self._control_mode = UnitType.XY
            return self._control_mode

    def SetControlMode(self, control_mode: UnitType or str):
        r"""
        Sets board control mode for this channel.

        Parameters
        ----------
        control_mode: UnitType or str
            Mode to set must be in register_tables.UnitType or appropriate string format given by
            UnitType enum attribute names.

        Raises
        ------
        none

        Returns
        -----
        str
            Firmware response or error information

        """
        if type(control_mode) is str:
            control_mode = UnitType.__members__[control_mode.upper()]

        if control_mode == UnitType.CURRENT and hasattr(self, 'FeedThrough'):
            response = self.FeedThrough.SetAsControlMode()
        elif control_mode == UnitType.OF and hasattr(self, 'OFPID'):
            response = self.OFPID.SetAsControlMode()
        elif control_mode == UnitType.XY and hasattr(self, 'XYPID'):
            response = self.XYPID.SetAsControlMode()
        else:
            return False

        self._control_mode = control_mode
        return response

    def initialize_channel(self):
        self.__init__(self._board, self._channel)

    # def _synchronize_channel(self):
    #     """
    #     Currently Unused. Intended for resetting register['value'], such that state could be stored in a log file and
    #     system returned to restore point.
    #     :return:
    #     """
    #     if self._board.verbose:
    #         print('Channel {} Synchronizing with firmware...'.format(self._channel))
    #     channel_systems = [system[1] for system in inspect.getmembers(self) if
    #                        type(system[1]) in systems()]
    #
    #     for i in range(len(channel_systems)):
    #         register_names = [name for name in channel_systems[i].__dict__ if not name.startswith('_')]
    #         for j in range(len(register_names)):
    #             register_name = register_names[j]
    #             register = channel_systems[i].__dict__[register_name]
    #             value = self._board.get_value(register)[2]
    #             if self._board.verbose:
    #                 print('Internally, {} {} set to {}'
    #                       .format(channel_systems[i].__class__.__name__, register_name, value))
    #
    #     if self._board.verbose:
    #         print('Channel {} Synchronized with firmware.'.format(self._channel))





__author__ = 'olgare'
__copyright__ = "Copyright 2021, Alpes Lasers SA"

import argparse
import os
import sys
from time import sleep

from flask import Flask
from flask.json import jsonify
from flask_restful import Api, Resource, abort
from flask_restful.reqparse import RequestParser

from phootonics_controller.base_controllers.config import WL_INCREMENT
from phootonics_controller.base_controllers.main_controller import MainController, AlreadyScanningException, \
    ControllerNotReady
import logging

logger = logging.getLogger(__name__)


def argsbool(value):
    if value == 't':
        return True
    elif value == 'f':
        return False
    else:
        raise ValueError('undefined value {}'.format(value))


app = Flask(__name__)

api = Api(app)


class ControllerResource(Resource):

    def __init__(self, controller, *args, **kwargs):
        self.controller = controller
        super(ControllerResource, self).__init__(*args, **kwargs)


class Info(ControllerResource):

    def get(self):
        monitoring_info = self.controller.get_monitoring_data()
        return jsonify({'monitoring': monitoring_info,
                        'scan_running': self.controller.is_scan_command_running(),
                        'scan': self.controller.get_results()})


def _pos_parser(x):
    return list(x)


scan_post_parser = RequestParser()
scan_post_parser.add_argument('positions', type=_pos_parser, required=True,
                              action='append',
                              help='list of (x, y) positions, example [(0.0, 0.0), (1.0, 0.0), ...]',
                              location=['json'])
scan_post_parser.add_argument('wl_step', type=float, required=False,
                              location=['json'],
                              default=WL_INCREMENT)


class ScanStart(ControllerResource):

    def post(self):
        args = scan_post_parser.parse_args()
        if self.controller.is_scan_command_running():
            abort(403, message='scan command already running')
        try:
            self.controller.start_scan(positions=args.positions, wl_step=args.wl_step)
        except AlreadyScanningException:
            abort(403, message='scan command already running')
        except ControllerNotReady:
            abort(403, message='controller not ready. Are all subsystems initialized?')


fast_scan_post_parser = RequestParser()
fast_scan_post_parser.add_argument('wl_step', type=float, required=False,
                                   location=['json'],
                                   default=WL_INCREMENT)
fast_scan_post_parser.add_argument('amplitude_x', type=float, required=False,
                                   location=['json'],
                                   default=1)
fast_scan_post_parser.add_argument('amplitude_y', type=float, required=False,
                                   location=['json'],
                                   default=1)


class FastScanStart(ControllerResource):

    def post(self):
        args = fast_scan_post_parser.parse_args()
        if self.controller.is_scan_command_running():
            abort(403, message='scan command already running')
        try:
            self.controller.start_fast_scan(wl_step=args.wl_step, amplitude_x=args.amplitude_x,
                                            amplitude_y=args.amplitude_y)
        except AlreadyScanningException:
            abort(403, message='scan command already running')
        except ControllerNotReady:
            abort(403, message='controller not ready. Are all subsystems initialized?')




class ScanStop(ControllerResource):

    def post(self):
        self.controller.stop_scan()


test_mode = (os.environ.get('PHOOTC_TEST_MODE', False) == 'MOCK')

controller = MainController(test_mode=test_mode)

try:
    controller.start()
except Exception as e:
    controller.shutdown_all_systems()
    raise


api.add_resource(Info, '/info',
                 resource_class_kwargs={'controller': controller})
api.add_resource(ScanStart, '/scan/start/default',
                 resource_class_kwargs={'controller': controller})
api.add_resource(FastScanStart, '/scan/start/fast',
                 resource_class_kwargs={'controller': controller})
api.add_resource(ScanStop, '/scan/stop',
                 resource_class_kwargs={'controller': controller})


def run():
    parser = argparse.ArgumentParser()
    parser.add_argument('-p', type=int, default=5555)
    args = parser.parse_args()
    try:
        app.run(host='0.0.0.0', port=args.p)
    finally:
        controller.shutdown_all_systems()


if __name__ == '__main__':
    run()


from .ClassAbstracts import System
from ..tools.systems_registers_tools import is_valid_channel


class OutputConditioning(System):
    r"""
Output Conditioning Channel - Offset and Scaling
System ID: 0xD0 through 0xdF

+---------------------+------+-------------+--------------------+-------------+---------+------------------------+
| Register Name       | Id   | Type        | Unit               | Range       | Default | Comment                |
+=====================+======+=============+====================+=============+=========+========================+
|                     |      |             |                    |             |         |                        |
+---------------------+------+-------------+--------------------+-------------+---------+------------------------+
    """

    # TODO: Not implemented yet in firmware
    @staticmethod
    def help():
        print(OutputConditioning.__doc__)

    def __init__(self, channel: int = 0, board=None):
        self.sys_id = 0xD0 | channel
        self._readonly = False

        self.gain = {'id': self.sys_id << 8 | 0x00,
                     'type': float,
                     'unit': None,
                     'range': None,
                     'default': 0,
                     'value': 0}
        self.offset = {'id': self.sys_id << 8 | 0x01,
                       'type': bool,
                       'unit': None,
                       'range': None,
                       'default': 1,
                       'value': 1}
        System.__init__(self, channel, board)
        self.name = self.__class__.__name__
        if not is_valid_channel(self._channel):
            raise ValueError('Channel Range Error')


from enum import IntEnum

# Kummenberg Internal Definitions
MAX_PAYLOAD_SIZE    = 50
SAMPLING_FREQ = 10000
ENDIAN = ">"
CHANNEL_CNT = 4
MAX_CHANNEL_CNT = 8
SIGNAL_FLOW_STAGE_CNT = 5
CMD_DELAY = 0.05
CHUNK_SIZE = 10
ESCAPE_BYTE = 0x7D
FRAME_BOUNDARY = 0x7E
ESCAPE_MASK = 1 << 5

SimpleResponses = {
    'PROCESSED_WITHOUT_ERROR' : b'OK\r\n',
    'OUT_OF_RANGE_UPPER'      : b'OU\r\n',
    'OUT_OF_RANGE_LOWER'      : b'OL\r\n',
    'UNRECOGNIZED_MESSAGE'    : b'NO\r\n',
    'ERROR'                   : b'ERROR\r\n'
}


class CommandID:
    GENERIC_COMMAND = 0x00
    GET_FIRMWARE_ID = 0X01
    GET_STATUS      = 0X02
    START_SELF_TEST = 0X03
    LOAD_CONFIG     = 0X04
    STORE_CONFIG    = 0X05
    SET_COMM_MODE   = 0X06
    SET_VALUE       = 0X10
    GET_VALUE       = 0X11
    SET_MULTIPLE    = 0X12
    GET_MULTIPLE    = 0X13
    SET_VECTOR      = 0X14
    GET_VECTOR      = 0X15

commandID = {
            'GENERIC_COMMAND' : 0X00,
            'GET_FIRMWARE_ID' : 0X01,
            'GET_STATUS'      : 0X02,
            'START_SELF_TEST' : 0X03,
            'LOAD_CONFIG'     : 0X04,
            'STORE_CONFIG'    : 0X05,
            'SET_COMM_MODE'   : 0X06,
            'SET_VALUE'       : 0X10,
            'GET_VALUE'       : 0X11,
            'SET_MULTIPLE'    : 0X12,
            'GET_MULTIPLE'    : 0X13,
            'SET_VECTOR'      : 0X14,
            'GET_VECTOR'      : 0X15
            }
commandName = dict((value, key) for key, value in commandID.items())


class UnitType(IntEnum):
    CURRENT   = 0
    OF        = 1
    XY     = 2
    UNDEFINED = 3


class Units:
    CURRENT     = 0
    OF          = 1
    XY          = 2
    UNDEFINED   = 3


class WaveformShape(IntEnum):
    SINUSOIDAL  = 0
    TRIANGULAR  = 1
    RECTANGULAR = 2
    SAWTOOTH    = 3
    PULSE       = 4


class Waveforms:
    SINUSOIDAL  = 0
    TRIANGULAR  = 1
    RECTANGULAR = 2
    SAWTOOTH    = 3
    PULSE       = 4


from phootonics_controller.utils.MR_E_2_Python_SDK.optoKummenberg.registers.ClassAbstracts import ControlStage
from phootonics_controller.utils.MR_E_2_Python_SDK.optoKummenberg.tools.systems_registers_tools import is_valid_channel


class CurrentFeedThrough(ControlStage):
    r"""
Control Mode Channel - Offset and Scaling
System ID: 0xB0 through 0xB7

+---------------------+------+-------------+--------------------+-------------+---------+------------------------+
| Register Name       | Id   | Type        | Unit               | Range       | Default | Comment                |
+=====================+======+=============+====================+=============+=========+========================+
|feed_through_enabled | 0x00 |             | None               |             | 0       |                        |
+---------------------+------+-------------+--------------------+-------------+---------+------------------------+

    """

    @staticmethod
    def help():
        print(FeedThrough.__doc__)

    def __init__(self, channel: int = 0, board=None):
        self.sys_id = 0xB0 | channel
        self._readonly = False

        self.feed_through_enabled = {'id': self.sys_id << 8 | 0x00,
                                     'type': bool,
                                     'unit': 'Hz',
                                     'range': [True, False],
                                     'default': 0,
                                     'value': 0}
        ControlStage.__init__(self, channel, board)
        self.name = self.__class__.__name__
        if not is_valid_channel(self._channel):
            raise ValueError('Channel Range Error')




# -*- coding: utf-8 -*-
import logging
import time
import math
import threading
from argparse import ArgumentParser

from phootonics_controller.utils.MR_E_2_Python_SDK.optoMDC.mre2 import MRE2Board
from phootonics_controller.utils.vcm_resources.base_controller import run_in_thread
from phootonics_controller.utils.vcm_resources.vcm_configuration import (configuration_dictionary, STAGE_ANGLE, update_configuration,
                                                         get_wavelength_for_angle)
from phootonics_controller.utils.vcm_resources.stage import StageBusyError, OutOfRangeError, StageController

logger = logging.getLogger(__name__)


def convert_x_position_to_angles_deg(pos):
    tan_50 = math.tan(math.radians(50))
    angle = (math.degrees(math.atan(pos * tan_50)))/2
    return angle


def convert_angles_deg_to_position_x(angle):
    tan_50 = math.tan(math.radians(50))
    pos = math.tan(math.radians(angle*2)) / tan_50
    return pos


class VCMirrorController(StageController):

    SIN = 0
    TRIANG = 1
    RECT = 2
    SAW = 3

    @property
    def is_busy(self):
        return self._device is not None and (self._busy_status not in [None, 'None', ''])

    def __init__(self, port):
        super(VCMirrorController, self).__init__(monitor_interval=0.1)
        self._port = port
        self._device_info = None
        self._busy_status = None
        self.filename = 'point_cloud.txt'
        self._device = None
        self._sweep_data = []
        self._stop_worker_event = threading.Event()
        self._has_two_axes = True

    def get_angle(self):
        return self._device.get_position()

    def get_angle_y(self):
        return self._device.get_position_y()

    def stop_motion(self):
        self._stop_worker_event.set()
        self._worker_thread = None

    def set_current_angle(self, angle):
        pass  # no-op for stages with homing capabilities

    def move_to_angle(self, angle):
        if self.is_busy:
            raise StageBusyError()
        self._device.move_abs(angle)

    def move_to_angle_y(self, angle_y):
        if self.is_busy:
            raise StageBusyError()
        self._device.move_y_axis_abs(angle_y)

    def jog(self, step):
        if self.is_busy:
            raise StageBusyError()
        self._device.set_step_size(abs(step))
        if step > 0:
            self._device.step_forward()
        else:
            self._device.step_backward()

    def sweep(self, amplitude, frequency, offset, shape):
        if self.is_busy:
            raise StageBusyError()
        self._stop_worker_event.clear()
        self._worker_thread = self._sweep_worker(amplitude, frequency, offset, shape)

    def stepwise(self, steps, delay_s):
        if self.is_busy:
            raise StageBusyError()
        self._stop_worker_event.clear()
        self._worker_thread = self._stepwise_worker(steps, delay_s)

    def convert_to_motiontrack(self, motion_track):
        conv_position = [(x[0], convert_x_position_to_angles_deg(x[1])) for x in motion_track]
        return [(x[0],x[1], get_wavelength_for_angle(x[1], False)) for x in conv_position]

    @run_in_thread
    def _sweep_worker(self, amplitude, frequency, offset, shape):
        self._sweep_data = []
        try:
            self._device.start_motion_track()
            self._busy_status = VCMirrorController.SWEEP
            self._device.set_signal_generator_parameters(amplitude, frequency, offset, shape)
            self._device.start_signal_generator()
            while not self._stop_worker_event.wait(1):
                gen_status = self._device.get_generator_status()
                if self._stop_worker_event.is_set():
                    self._device.stop_signal_generator()
                    break
                if gen_status == 0:
                    self._device.stop_signal_generator()
                    break
        except Exception as e:
            logger.error('Sweep error {}'.format(e))
            raise
        finally:
            self._device.stop_motion_track()
            self._device.stop_signal_generator()
        current_sweep = self.convert_to_motiontrack(self._device.get_motion_track())
        self._sweep_data = [current_sweep]
        self._busy_status = None

    @run_in_thread
    def _stepwise_worker(self, steps, delay_s):
        try:
            self._device.start_motion_track()
            self._busy_status = VCMirrorController.STEPWISE
            for step in steps:
                if self._stop_worker_event.is_set():
                    break
                self._device.move_abs(step)
                time.sleep(delay_s)
        finally:
            self._device.stop_motion_track()
        current_step = self.convert_to_motiontrack(self._device.get_motion_track())
        self._sweep_data = [current_step]
        self._busy_status = None

    def _connected_status(self):
        if self.is_busy:
            return self._busy_status or VCMirrorController.BUSY
        return VCMirrorController.READY

    def _disconnect(self):
        try:
            self.stop_motion()
        finally:
            try:
                self._device.close()
            except Exception as e:
                logger.info('_disconnect error {}'.format(e), exc_info=1)
                self._device = None

    def _connect(self):
        self._device = VCMirror(self._port)
        self._device.open()
        self._device_info = {'serial_number': self._device.serial_number, 'model': 'MR-E-2',
                             'firmware_version': self._device.sw_verison}

    def _check_connection(self):
        if self._device:
            last_angle = self.get_angle()
            if last_angle is not None:
                if last_angle != 'N/A':
                    update_configuration({STAGE_ANGLE: last_angle})
            return self._device.check_connection()
        else:
            logger.error('self._device of VCMirrorController is None')
            return False


class VCMirror:

    @property
    def serial_number(self):
        return self._serial_number

    @property
    def sw_verison(self):
        return self._sw_version

    def __init__(self, port, loop_delay=0.001):
        self.data_lock = threading.RLock()
        self.stopEvent = threading.Event()
        self.thread = None
        self._serial_number = ''
        self._sw_version = ''
        self._port = port
        self.device = None
        self.loop_delay = loop_delay
        self._position = None
        self._position_y = None
        self._temperature = None
        self._last_recording = None
        self._step_size = 0.2
        self.sig_gen_params_valid = False
        self._is_connected = False
        self._motion_track = None
        self._is_recording = False
        self.record_callback = lambda x, y, t: None

    def register_record_callback(self, f):
        self.record_callback = f

    def unregister_record_callback(self, f):
        self.record_callback = lambda x, y, t: None

    def connection_ok(self):
        self._is_connected = True

    def connection_ko(self):
        self._is_connected = False

    def start_motion_track(self):
        with self.data_lock:
            self._motion_track = []
            self._is_recording = True

    def stop_motion_track(self):
        self._is_recording = False

    def get_motion_track(self):
        return self._motion_track

    def _record_motion(self):
        if self._is_recording:
            self._motion_track.append((self._last_recording,
                                       self._position))

    def _home(self):
        with self.data_lock:
            self.device.Mirror.Channel_0.StaticInput.SetXY(0.0)
            self.device.Mirror.Channel_1.StaticInput.SetXY(0.0)

    def _set_xy_control_mode_both_axes(self):
        with self.data_lock:
            current_control_mode_x = self.device.Mirror.Channel_0.GetControlMode()
            if current_control_mode_x != 2:
                self.device.Mirror.Channel_0.SetControlMode(2)
            current_control_mode_y = self.device.Mirror.Channel_1.GetControlMode()
            if current_control_mode_y != 2:
                self.device.Mirror.Channel_1.SetControlMode(2)

    def _init(self):
        self._set_xy_control_mode_both_axes()
        self._home()
        self.move_y_axis_abs(configuration_dictionary['STAGE_ANGLE_Y'])
        with self.data_lock:
            self._serial_number = self.device.Status.get_register('cpu_fpga_version')[0]
            self._sw_version = self.device.Status.get_register('firmware_id')[0]

    def open(self):
        if not self.device:
            self.device = MRE2Board(port=self._port, verbose=True)
        self._init()
        self.connection_ok()
        self.thread = threading.Thread(target=self._run)
        self.thread.start()

    def close(self):
        try:
            self.stopEvent.set()
            self.thread.join(2)
            if self.thread.isAlive():
                logger.error('stage vcm _run thread is still alive on close')
            self.device.disconnect()
        finally:
            self.thread = None
            self.device = None
            self.connection_ko()

    def check_connection(self):
        if not self.thread or (not self.thread.isAlive()):
            return False
        return self._is_connected

    def _run(self):
        while self.stopEvent.wait(self.loop_delay) is False:
            if self.stopEvent.is_set():
                return
            try:
                with self.data_lock:
                    pos_x_cal = self.device.Mirror.OpticalFeedback.get_register('x')[0]
                    pos_y_cal = self.device.Mirror.OpticalFeedback.get_register('y')[0]
                    pos_angle = convert_x_position_to_angles_deg(pos_x_cal)
                    pos_angle_y = convert_x_position_to_angles_deg(pos_y_cal)
                    self._position = pos_angle
                    self._position_y = pos_angle_y
                    self._temperature = self.device.TemperatureManager.GetProxyTemperature()[0]
                    self._last_recording = time.time()
                    self.record_callback(pos_angle, pos_angle_y, self._last_recording)
            except Exception as e:
                logger.exception(e)
                logger.info('error communicating with VCM')
                return

    def get_position(self):
        return self._position

    def get_position_y(self):
        return self._position_y

    def move_abs(self, angle):
        with self.data_lock:
            pos = convert_angles_deg_to_position_x(angle)
            if self.angle_not_out_of_bounds(angle):
                self.device.Mirror.Channel_0.StaticInput.SetXY(pos)
            else:
                raise OutOfRangeError()

    def move_y_axis_abs(self, angle):
        with self.data_lock:
            posy = convert_angles_deg_to_position_x(angle)
            if self.angle_not_out_of_bounds(angle):
                self.device.Mirror.Channel_1.StaticInput.SetXY(posy)
            else:
                raise OutOfRangeError()

    def move_rel(self, increment):
        with self.data_lock:
            current_position_x = float(self.device.Mirror.Channel_0.StaticInput.GetXY()[0])
            current_angle_x = convert_x_position_to_angles_deg(current_position_x)
            target_angle_x = current_angle_x + increment
            target_position_x = convert_angles_deg_to_position_x(target_angle_x)
            if self.angle_not_out_of_bounds(target_angle_x):
                self.device.Mirror.Channel_0.StaticInput.SetXY(target_position_x)
            else:
                raise OutOfRangeError()

    def step_forward(self):
        self.move_rel(self._step_size)

    def step_backward(self):
        self.move_rel(self._step_size*(-1))

    def get_step_size(self):
        return self._step_size

    def set_step_size(self, step_size):
        self._step_size = abs(step_size)

    def angle_not_out_of_bounds(self, position):
        if position >-25 and position < 25 :
            return True
        else:
            return False

    def set_signal_generator_parameters(self, amplitude, frequency, offset, shape):
        with self.data_lock:
            amp = convert_angles_deg_to_position_x(amplitude)
            off = convert_angles_deg_to_position_x(offset)
            max_position = abs(amplitude) + abs(offset)
            if self.angle_not_out_of_bounds(max_position):
                self.sig_gen_params_valid = True
            else:
                self.sig_gen_params_valid = False
            if self.sig_gen_params_valid:
                self.device.Mirror.Channel_0.SignalGenerator.SetAmplitude(amp)
                self.device.Mirror.Channel_0.SignalGenerator.SetFrequency(frequency)
                self.device.Mirror.Channel_0.SignalGenerator.SetAsInput()
                self.device.Mirror.Channel_0.SignalGenerator.SetUnit(2)
                self.device.Mirror.Channel_0.SignalGenerator.SetShape(shape)
                self.device.Mirror.Channel_0.SignalGenerator.SetCycles(-1)
                self.device.Mirror.Channel_0.SignalGenerator.SetOffset(off)
            else:
                raise OutOfRangeError()

    def start_x_signal_generator(self, amplitude, frequency, offset, shape):
        with self.data_lock:
            amp = convert_angles_deg_to_position_x(amplitude)
            off = convert_angles_deg_to_position_x(offset)
            max_position = abs(amplitude) + abs(offset)
            valid =  self.angle_not_out_of_bounds(max_position)

            if valid:
                self.device.Mirror.Channel_0.SignalGenerator.SetAmplitude(amp)
                self.device.Mirror.Channel_0.SignalGenerator.SetFrequency(frequency)
                self.device.Mirror.Channel_0.SignalGenerator.SetAsInput()
                self.device.Mirror.Channel_0.SignalGenerator.SetUnit(2)
                self.device.Mirror.Channel_0.SignalGenerator.SetShape(shape)
                self.device.Mirror.Channel_0.SignalGenerator.SetCycles(-1)
                self.device.Mirror.Channel_0.SignalGenerator.SetOffset(off)
                self.device.Mirror.Channel_0.SignalGenerator.Run()
            else:
                raise OutOfRangeError

    def start_y_signal_generator(self, amplitude, frequency, offset, shape):
        with self.data_lock:
            amp = convert_angles_deg_to_position_x(amplitude)
            off = convert_angles_deg_to_position_x(offset)
            max_position = abs(amplitude) + abs(offset)
            valid =  self.angle_not_out_of_bounds(max_position)
            if valid:
                self.device.Mirror.Channel_1.SignalGenerator.SetAmplitude(amp)
                self.device.Mirror.Channel_1.SignalGenerator.SetFrequency(frequency)
                self.device.Mirror.Channel_1.SignalGenerator.SetAsInput()
                self.device.Mirror.Channel_1.SignalGenerator.SetUnit(2)
                self.device.Mirror.Channel_1.SignalGenerator.SetShape(shape)
                self.device.Mirror.Channel_1.SignalGenerator.SetCycles(-1)
                self.device.Mirror.Channel_1.SignalGenerator.SetOffset(off)
                self.device.Mirror.Channel_1.SignalGenerator.Run()
            else:
                raise OutOfRangeError

    def reset_static_input(self):
        with self.data_lock:
            self.device.Mirror.Channel_0.SignalGenerator.Stop()
            self.device.Mirror.Channel_1.SignalGenerator.Stop()
            self.device.Mirror.Channel_0.StaticInput.SetAsInput()
            self.device.Mirror.Channel_1.StaticInput.SetAsInput()

    def start_signal_generator(self):
        with self.data_lock:
            if self.sig_gen_params_valid:
                self.device.Mirror.Channel_0.SignalGenerator.Run()
            else:
                raise OutOfRangeError()

    def get_generator_status(self):
        with self.data_lock:
            gen_status= self.device.Mirror.Channel_0.SignalGenerator.GetRunningStatus()[0]
            return gen_status

    def stop_signal_generator(self):
        with self.data_lock:
            self.device.Mirror.Channel_0.SignalGenerator.Stop()
            self.set_static_input()

    def set_static_input(self):
        with self.data_lock:
            self.device.Mirror.Channel_0.StaticInput.SetAsInput()

def move():
    parser = ArgumentParser()
    parser.add_argument('position_x', type=float)
    parser.add_argument('position_y', type=float)
    parser.add_argument('-p', type=str)
    args=parser.parse_args()
    d = VCMirror(args.p)
    d.open()
    try:
        d.move_abs(args.position_x)
        d.move_y_axis_abs(args.position_y)
    finally:
        d.close()


if __name__ == '__main__':
    move()





r"""
Lookup tables / management of System registers.

Each System is represented by its own class, with Registers as attributes of that class.
The registers module can be imported to retrieve information about a given registers.
Syntax is as follows: registers.SYSTEM_NAME(CHANNEL_NUMBER).REGISTER_NAME

Returns
-------
dict
    A standard lookup will return a Register Record, or dictionary, containing pertinent registers information.

Notes
-----
A handful of helper methods are included to manage registers requests. See corresponding info for the following:
    is_valid_channel(value), is_valid_value(register_dict, value)
    encode(command_id. register_id)
    system_info()
    system_names()
    register_names(system_name)
    get_register_id(system_name, register_name, _channel)
    parse_error(error_code)
    help()

Examples
--------

For example, this imports the registers module, and collects the registers record of USB channel 0, as well as the
registers record for setting the input channel system of the Signal Flow Manager.

    >>> import optoKummenberg
    >>> optoKummenberg.Registers.StaticInput(0).of

    >>> reg.StaticInput(0).of
    {'id': 20737, 'type': <class 'float'>, 'unit': None, 'range': [-1, 1], 'default': 0.0}
    >>> reg.Manager(0).input
    {'id': 16384, 'type': <class 'int'>, 'unit': 'SystemID', 'range': None, 'default': None}
    >>> optoKummenberg.tools.systems_registers_tools.is_valid_value(registers.StaticInput.angle, 22.1)
    True

Additional Notes
----------------
For more info on a given registers, see the help text in the corresponding System class.
This will display ID, Data Type, Units, Range, Default Value, and Comments.
    >>> registers.SignalGenerator.help()

For a complete list of Systems, call one of the helper methods from registers module itself.
    >>> optoMDC.tools.systems_registers_tools.system_names()
    >>> optoMDC.tools.systems_registers_tools.system_info()

"""
# other
import inspect
import sys
from ..tools.definitions import CHANNEL_CNT
from ..tools.systems_registers_tools import get_registers
# input stage
from .InputStage import *
# input conditoning
from .InputConditioning import *
# control mode
from .ControlStage import *
# output conditioning
from .OutputConditioning import *
# output
from .OutputStage import *
# misc systems
from .MiscSystems import *


def help():
    print(__doc__)


def systems():
    all_systems = inspect.getmembers(sys.modules[__name__], inspect.isclass)
    all_systems = list(zip(*all_systems))[1]
    sys_dict = {}
    for system in all_systems:
        try:
            for i in range(CHANNEL_CNT):
                sys_obj = system(channel=i)
                sys_id = sys_obj.sys_id
                reg_dict = dict(sys_obj.register_list)
                sys_dict.update({sys_id: {'name': sys_obj.name, 'registers': reg_dict}})
        except TypeError:
            # not a system
            pass
        except AttributeError:
            # abstract system, do not add to list
            pass
    return sys_dict
    # return [system for system in all_systems if system()._is_a_system]
